"""Tests `imitation.util.networks`."""

import functools
import math
from typing import Type

import pytest
import torch as th

from imitation.util import networks

assert_equal = functools.partial(th.testing.assert_close, rtol=0, atol=0)


NORMALIZATION_LAYERS = [
    networks.RunningNorm,
    networks.EMANorm,
]


class EMANormAlgorithm2(networks.EMANorm):
    """EMA Norm using algorithm 2 from the reference below.

    Reference:
    https://github.com/HumanCompatibleAI/imitation/files/9456540/Incremental_batch_EMA_and_EMV.pdf
    """

    def __init__(
        self,
        num_features: int,
        decay: float = 0.99,
        eps: float = 1e-5,
    ):
        """Builds EMARunningNormIncremental."""
        super().__init__(
            num_features,
            decay=decay,
            eps=eps,
        )

    def update_stats(self, batch: th.Tensor) -> None:
        """Update `self.running_mean` and `self.running_var` incrementally.

        Reference Algorithm 2 from:
        https://github.com/HumanCompatibleAI/imitation/files/9364938/Incremental_batch_EMA_and_EMV.pdf

        Args:
            batch: A batch of data to use to update the running mean and variance.
        """
        b_size = batch.shape[0]
        if len(batch.shape) == 1:
            batch = batch.reshape(b_size, 1)

        self.inv_learning_rate += self.decay**self.num_batches
        learning_rate = 1 / self.inv_learning_rate
        if self.count == 0:
            self.running_mean = batch.mean(dim=0)
            if b_size > 1:
                self.running_var = batch.var(dim=0, unbiased=False)
            else:
                self.running_var = th.zeros_like(self.running_mean, dtype=th.float)
        else:
            S = th.mean((batch - self.running_mean) ** 2, dim=0)

            # update running mean
            delta = learning_rate * (batch.mean(0) - self.running_mean)
            self.running_mean += delta

            # update running variance
            self.running_var *= 1 - learning_rate
            self.running_var += learning_rate * S - delta**2

        self.count += b_size
        self.num_batches += 1  # type: ignore[misc]


@pytest.mark.parametrize("normalization_layer", NORMALIZATION_LAYERS)
def test_running_norm_identity_eval(normalization_layer: Type[networks.BaseNorm]):
    """Tests running norm starts and stays at identity function when in eval mode.

    Specifically, we test in evaluation mode (initialization should not change)

    Args:
        normalization_layer: the normalization layer to be tested.
    """
    running_norm = normalization_layer(1, eps=0.0)
    x = th.Tensor([-1.0, 0.0, 7.32, 42.0])
    running_norm.eval()  # stats should not change in eval mode
    for _ in range(10):
        assert_equal(running_norm.forward(x), x)


@pytest.mark.parametrize("normalization_layer", NORMALIZATION_LAYERS)
def test_running_norm_identity_train(normalization_layer: Type[networks.BaseNorm]):
    """Test that the running norm will not change already normalized data.

    Args:
        normalization_layer: the normalization layer to be tested.
    """
    running_norm = normalization_layer(1, eps=0.0)
    running_norm.train()  # stats will change in eval mode
    normalized = th.Tensor([-1, -1, -1, -1, 1, 1, 1, 1])  # mean 0, variance 1
    for _ in range(10):
        th.testing.assert_close(
            running_norm.forward(normalized),
            normalized,
            rtol=0.05,
            atol=0.05,
        )


@pytest.mark.parametrize("normalization_layer", NORMALIZATION_LAYERS)
def test_running_norm_eval_fixed(
    normalization_layer: Type[networks.BaseNorm],
    batch_size: int = 8,
    num_batches: int = 10,
    num_features: int = 4,
) -> None:
    """Tests that stats do not change when in eval mode and do when in training."""
    running_norm = normalization_layer(num_features)

    def do_forward(shift: float = 0.0, scale: float = 1.0):
        for i in range(num_batches):
            data = th.rand(batch_size, num_features) * scale + shift
            running_norm.forward(data)

    with th.random.fork_rng():
        th.random.manual_seed(42)

        do_forward()
        current_mean = th.clone(running_norm.running_mean)
        current_var = th.clone(running_norm.running_var)

        running_norm.eval()
        do_forward()
        assert_equal(running_norm.running_mean, current_mean)
        assert_equal(running_norm.running_var, current_var)

        running_norm.train()
        do_forward(1.0, 2.0)
        assert th.all((running_norm.running_mean - current_mean).abs() > 0.01)
        assert th.all((running_norm.running_var - current_var).abs() > 0.01)


@pytest.mark.parametrize("batch_size", [1, 8])
def test_running_norm_matches_dist(batch_size: int) -> None:
    """Test running norm converges to empirical distribution."""
    mean = th.Tensor([-1.3, 0.0, 42])
    var = th.Tensor([0.1, 1.0, 42])
    sd = th.sqrt(var)

    num_dims = len(mean)
    running_norm = networks.RunningNorm(num_dims)
    running_norm.train()

    num_samples = 256
    with th.random.fork_rng():
        th.random.manual_seed(42)
        data = th.randn(num_samples, num_dims) * sd + mean
        for start in range(0, num_samples, batch_size):
            batch = data[start : start + batch_size]
            running_norm.forward(batch)

    empirical_mean = th.mean(data, dim=0)
    empirical_var = th.var(data, dim=0, unbiased=False)

    normalized = th.Tensor([[-1.0], [0.0], [1.0], [42.0]])
    normalized = th.tile(normalized, (1, 3))
    scaled = normalized * th.sqrt(empirical_var + running_norm.eps) + empirical_mean
    running_norm.eval()
    for i in range(5):
        th.testing.assert_close(running_norm.forward(scaled), normalized)

    # Stats should match empirical mean (and be unchanged by eval)
    th.testing.assert_close(running_norm.running_mean, empirical_mean)
    th.testing.assert_close(running_norm.running_var, empirical_var)
    assert running_norm.count == num_samples


@pytest.mark.parametrize("batch_size", [1, 8])
@pytest.mark.parametrize("normalization_layer", NORMALIZATION_LAYERS)
def test_parameters_converge(
    batch_size: int,
    normalization_layer: Type[networks.BaseNorm],
) -> None:
    """Test running norm parameters approximately converge to true values."""
    mean = th.Tensor([3, 0])
    var = th.Tensor([6, 1])
    sd = th.sqrt(var)

    num_dims = len(mean)
    running_norm = normalization_layer(num_dims)
    running_norm.train()

    num_samples = 2000
    with th.random.fork_rng():
        th.random.manual_seed(42)
        data = th.randn(num_samples, num_dims) * sd + mean
        for start in range(0, num_samples, batch_size):
            batch = data[start : start + batch_size]
            running_norm.forward(batch)

    running_norm.eval()
    th.testing.assert_close(running_norm.running_mean, mean, rtol=0.05, atol=0.1)
    th.testing.assert_close(running_norm.running_var, var, rtol=0.1, atol=0.1)

    assert running_norm.count == num_samples


@pytest.mark.parametrize(
    "init_kwargs",
    [{}, {"dropout_prob": 0.5}]
    + [{"normalize_input_layer": layer} for layer in NORMALIZATION_LAYERS],
)
def test_build_mlp_norm_training(init_kwargs) -> None:
    """Tests MLP building function `networks.build_mlp()`.

    Specifically, we initialize an MLP and train it on a toy task. We also test the
    init options of input layer normalization.

    Args:
        init_kwargs: dict of kwargs to pass to `networks.build_mlp()`.
    """
    # Create Tensors to hold input and outputs.
    x = th.linspace(-math.pi, math.pi, 200).reshape(-1, 1)
    y = th.sin(x)
    # Construct our model by instantiating the class defined above
    model = networks.build_mlp(in_size=1, hid_sizes=[16, 16], out_size=1, **init_kwargs)

    # Construct a loss function and an Optimizer.
    criterion = th.nn.MSELoss(reduction="sum")
    optimizer = th.optim.SGD(model.parameters(), lr=1e-6)
    for t in range(200):
        # Forward pass: Compute predicted y by passing x to the model
        y_pred = model(x)

        # Compute and print loss
        loss = criterion(y_pred, y)

        # Zero gradients, perform a backward pass, and update the weights.
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()


def test_build_mlp_raises_on_invalid_normalize_input_layer() -> None:
    """Test that `networks.build_mlp()` raises on invalid input layer."""
    with pytest.raises(
        ValueError,
        match="normalize_input_layer.*not a valid normalization layer.*",
    ):
        networks.build_mlp(
            in_size=1,
            hid_sizes=[16, 16],
            out_size=1,
            normalize_input_layer=th.nn.Module,
        )


def test_input_validation_on_ema_norm():
    with pytest.raises(ValueError):
        networks.EMANorm(128, decay=1.1)

    with pytest.raises(ValueError):
        networks.EMANorm(128, decay=-0.1)

    networks.EMANorm(128, decay=0.05)


@pytest.mark.parametrize("decay", [0.5, 0.99])
@pytest.mark.parametrize("input_shape", [(64,), (1, 256), (64, 256)])
def test_ema_norm_algo_2_and_3_are_the_same(decay, input_shape):
    num_features = input_shape[-1] if len(input_shape) == 2 else 1
    ema_norm_algo_2 = EMANormAlgorithm2(
        num_features=num_features,
        decay=decay,
    )
    ema_norm_algo_3 = networks.EMANorm(num_features, decay)
    random_tensor = th.randn(input_shape)
    for i in range(1000):
        ema_norm_algo_2.train(), ema_norm_algo_3.train()
        # moving distribution
        tensor_input = random_tensor.clone() + i
        ema_norm_algo_2(tensor_input)
        tensor_input = random_tensor.clone() + i
        ema_norm_algo_3(tensor_input)

        ema_norm_algo_2.eval(), ema_norm_algo_3.eval()
        th.testing.assert_close(
            ema_norm_algo_2.running_mean,
            ema_norm_algo_3.running_mean,
            rtol=0.05,
            atol=0.1,
        )
        th.testing.assert_close(
            ema_norm_algo_2.running_var,
            ema_norm_algo_3.running_var,
            rtol=0.05,
            atol=0.1,
        )



"""This ingredient performs evaluation of learned policy.

It takes care of the right wrappers, does some rollouts
and computes statistics of the rollouts.
"""

from typing import Mapping, Union

import numpy as np
import sacred
from stable_baselines3.common import base_class, policies, vec_env

from imitation.data import rollout

policy_evaluation_ingredient = sacred.Ingredient("policy_evaluation")


@policy_evaluation_ingredient.config
def config():
    n_episodes_eval = 50  # Num of episodes for final mean ground truth return
    locals()  # quieten flake8


@policy_evaluation_ingredient.named_config
def fast():
    n_episodes_eval = 1  # noqa: F841


@policy_evaluation_ingredient.capture
def eval_policy(
    rl_algo: Union[base_class.BaseAlgorithm, policies.BasePolicy],
    venv: vec_env.VecEnv,
    n_episodes_eval: int,
    _rnd: np.random.Generator,
) -> Mapping[str, float]:
    """Evaluation of imitation learned policy.

    Has the side effect of setting `rl_algo`'s environment to `venv`
    if it is a `BaseAlgorithm`.

    Args:
        rl_algo: Algorithm to evaluate.
        venv: Environment to evaluate on.
        n_episodes_eval: The number of episodes to average over when calculating
            the average episode reward of the imitation policy for return.
        _rnd: Random number generator provided by Sacred.

    Returns:
        A dictionary with two keys. "imit_stats" gives the return value of
        `rollout_stats()` on rollouts test-reward-wrapped environment, using the final
        policy (remember that the ground-truth reward can be recovered from the
        "monitor_return" key). "expert_stats" gives the return value of
        `rollout_stats()` on the expert demonstrations loaded from `path`.
    """
    sample_until_eval = rollout.make_min_episodes(n_episodes_eval)
    if isinstance(rl_algo, base_class.BaseAlgorithm):
        # Set RL algorithm's env to venv, removing any cruft wrappers that the RL
        # algorithm's environment may have accumulated.
        rl_algo.set_env(venv)
        # Generate trajectories with the RL algorithm's env - SB3 may apply wrappers
        # under the hood to get it to work with the RL algorithm (e.g. transposing
        # images, so they can be fed into CNNs).
        train_env = rl_algo.get_env()
        assert train_env is not None
    else:
        train_env = venv
    trajs = rollout.generate_trajectories(
        rl_algo,
        train_env,
        sample_until=sample_until_eval,
        rng=_rnd,
    )
    return rollout.rollout_stats(trajs)



"""Tests of `imitation.data.types`."""

import contextlib
import copy
import dataclasses
import os
import pathlib
import pickle
from typing import Any, Callable, Sequence

import gymnasium as gym
import numpy as np
import pytest

from imitation.data import serialize, types
from imitation.util import util

SPACES = [
    gym.spaces.Discrete(3),
    gym.spaces.MultiDiscrete([3, 4]),
    gym.spaces.Box(-1, 1, shape=(1,)),
    gym.spaces.Box(-1, 1, shape=(2,)),
    gym.spaces.Box(-np.inf, np.inf, shape=(2,)),
]
DICT_SPACE = gym.spaces.Dict(
    {"a": gym.spaces.Discrete(3), "b": gym.spaces.Box(-1, 1, shape=(2,))},
)

OBS_SPACES = SPACES + [DICT_SPACE]
ACT_SPACES = SPACES
LENGTHS = [0, 1, 2, 10]


def _check_1d_shape(fn: Callable[[np.ndarray], Any], length: int, expected_msg: str):
    for shape in [(), (length, 1), (length, 2), (length - 1,), (length + 1,)]:
        with pytest.raises(ValueError, match=expected_msg):
            fn(np.zeros(shape))


@pytest.fixture
def trajectory(
    obs_space: gym.Space,
    act_space: gym.Space,
    length: int,
) -> types.Trajectory:
    """Fixture to generate trajectory of length `length` iid sampled from spaces."""
    if length == 0:
        pytest.skip()

    raw_obs = [obs_space.sample() for _ in range(length + 1)]
    if isinstance(obs_space, gym.spaces.Dict):
        obs: types.Observation = types.DictObs.from_obs_list(raw_obs)
    else:
        obs = np.array(raw_obs)
    acts = np.array([act_space.sample() for _ in range(length)])
    infos = np.array([{f"key{i}": i} for i in range(length)])
    return types.Trajectory(obs=obs, acts=acts, infos=infos, terminal=True)


@pytest.fixture
def trajectory_rew(trajectory: types.Trajectory) -> types.TrajectoryWithRew:
    """Like `trajectory` but with reward randomly sampled from a Gaussian."""
    rews = np.random.randn(len(trajectory))
    return types.TrajectoryWithRew(
        **types.dataclass_quick_asdict(trajectory),
        rews=rews,
    )


@pytest.fixture
def transitions_min(
    obs_space: gym.Space,
    act_space: gym.Space,
    length: int,
) -> types.TransitionsMinimal:
    obs = np.array([obs_space.sample() for _ in range(length)])
    acts = np.array([act_space.sample() for _ in range(length)])
    infos = np.array([{i: i} for i in range(length)])
    return types.TransitionsMinimal(obs=obs, acts=acts, infos=infos)


@pytest.fixture
def transitions(
    transitions_min: types.TransitionsMinimal,
    obs_space: gym.Space,
    length: int,
) -> types.Transitions:
    """Fixture to generate transitions of length `length` iid sampled from spaces."""
    next_obs = np.array([obs_space.sample() for _ in range(length)])
    dones = np.zeros(length, dtype=bool)
    return types.Transitions(
        **types.dataclass_quick_asdict(transitions_min),
        next_obs=next_obs,
        dones=dones,
    )


@pytest.fixture
def transitions_rew(
    transitions: types.Transitions,
    length: int,
) -> types.TransitionsWithRew:
    """Like `transitions` but with reward randomly sampled from a Gaussian."""
    rews = np.random.randn(length)
    return types.TransitionsWithRew(
        **types.dataclass_quick_asdict(transitions),
        rews=rews,
    )


def _check_transitions_get_item(trans, key):
    """Check trans[key] by manually indexing/slicing into every `trans` field."""
    item = trans[key]
    for field in dataclasses.fields(trans):
        if isinstance(item, dict):
            observed = item[field.name]  # pytype: disable=unsupported-operands
        else:
            observed = getattr(item, field.name)

        expected = getattr(trans, field.name)[key]
        if isinstance(expected, np.ndarray):
            assert observed.dtype == expected.dtype  # pytype:disable=attribute-error
        np.testing.assert_array_equal(observed, expected)


@contextlib.contextmanager
def pushd(dir_path):
    """Change directory temporarily inside context."""
    orig_dir = pathlib.Path.cwd()
    try:
        os.chdir(dir_path)
        yield
    finally:
        os.chdir(orig_dir)


@pytest.mark.parametrize("obs_space", OBS_SPACES)
@pytest.mark.parametrize("act_space", ACT_SPACES)
@pytest.mark.parametrize("length", LENGTHS)
class TestData:
    """Tests of imitation.util.data.

    Grouped in a class since parametrized over common set of spaces.
    """

    def test_valid_trajectories(
        self,
        trajectory: types.Trajectory,
        trajectory_rew: types.TrajectoryWithRew,
        length: int,
    ) -> None:
        """Checks trajectories can be created for a variety of lengths and spaces."""
        trajs = [trajectory, trajectory_rew]
        trajs += [dataclasses.replace(traj, infos=None) for traj in trajs]
        for traj in trajs:
            assert len(traj) == length

    def test_traj_unequal_to_other_types(
        self,
        trajectory: types.Trajectory,
        trajectory_rew: types.TrajectoryWithRew,
    ) -> None:
        """Test trajectories unequal to objects of different types."""
        for t in [trajectory, trajectory_rew]:
            # Trajectory compare unequal to things that are not trajectories
            assert t != 42
            assert t != "foobar"

        # Trajectory compares unequal to a copy of itself but with reward
        assert trajectory != trajectory_rew

    def test_traj_equal_to_self_and_copies(
        self,
        trajectory: types.Trajectory,
        trajectory_rew: types.TrajectoryWithRew,
    ) -> None:
        """Test that trajectories are equal to themselves and copies."""
        for t in [trajectory, trajectory_rew]:
            # Equal to itself
            assert t == t
            # And to copy
            assert t == copy.copy(t)

    def test_traj_unequal_to_perturbations(
        self,
        trajectory: types.Trajectory,
        trajectory_rew: types.TrajectoryWithRew,
        length: int,
    ) -> None:
        """Test that trajectories unequal to perturbed versions."""
        # Unequal to a copy of itself truncated
        new_length = length - 1
        if new_length > 0:
            assert trajectory != types.Trajectory(
                obs=trajectory.obs[: new_length + 1],
                acts=trajectory.acts[:new_length],
                infos=trajectory.infos[:new_length]
                if trajectory.infos is not None
                else None,
                terminal=trajectory.terminal,
            )

        # Or with contents changed
        for t in [trajectory, trajectory_rew]:
            as_dict = types.dataclass_quick_asdict(t)
            for k in as_dict.keys():
                perturbed = dict(as_dict)
                if k == "infos":
                    perturbed["infos"] = [{"foo": 42}] * len(as_dict["infos"])
                elif isinstance(as_dict[k], types.DictObs):
                    perturbed[k] = as_dict[k].map_arrays(lambda x: x + 1)
                else:
                    perturbed[k] = as_dict[k] + 1
                assert t != type(t)(**perturbed)

    @pytest.mark.parametrize("type_safe", [False, True])
    @pytest.mark.parametrize("use_pickle", [False, True])
    @pytest.mark.parametrize("use_rewards", [False, True])
    @pytest.mark.parametrize("use_chdir", [False, True])
    def test_save_trajectories(
        self,
        trajectory: types.Trajectory,
        trajectory_rew: types.TrajectoryWithRew,
        use_chdir,
        tmpdir,
        use_pickle,
        use_rewards,
        type_safe,
    ):
        if isinstance(trajectory.obs, types.DictObs):
            pytest.xfail("Saving/loading dictobs trajectories not yet supported")

        chdir_context: contextlib.AbstractContextManager
        """Check that trajectories are properly saved."""
        if use_chdir:
            # Test no relative path without directory edge-case.
            chdir_context = pushd(tmpdir)
            save_dir_str = ""
        else:
            chdir_context = contextlib.nullcontext()
            save_dir_str = tmpdir

        with chdir_context:
            save_dir = util.parse_path(save_dir_str)
            trajs = [trajectory_rew if use_rewards else trajectory]
            save_path = save_dir / "trajs"

            if use_pickle:
                # Pickle format
                with open(save_path, "wb") as f:
                    pickle.dump(trajs, f)
            else:
                # HuggingFace Dataset Format
                serialize.save(save_path, trajs)

                # Test that heterogeneous lists of trajectories throw an error
                if use_rewards:
                    with pytest.raises(ValueError):
                        serialize.save(save_path, [trajectory, trajectory_rew])

            loaded_trajs: Sequence[types.Trajectory]
            if type_safe:
                if use_rewards:
                    loaded_trajs = serialize.load_with_rewards(save_path)
                else:
                    with pytest.raises(ValueError):
                        serialize.load_with_rewards(save_path)
                    loaded_trajs = serialize.load(save_path)
            else:
                loaded_trajs = serialize.load(save_path)

            assert len(trajs) == len(loaded_trajs)
            for t1, t2 in zip(trajs, loaded_trajs):
                assert t1 == t2

    def test_invalid_trajectories(
        self,
        trajectory: types.Trajectory,
        trajectory_rew: types.TrajectoryWithRew,
    ) -> None:
        """Checks input validation catches space and dtype related errors."""
        trajs = [trajectory, trajectory_rew]
        for traj in trajs:
            with pytest.raises(
                ValueError,
                match=r"expected one more observations than actions.*",
            ):
                dataclasses.replace(traj, obs=traj.obs[:-1])
            with pytest.raises(
                ValueError,
                match=r"expected one more observations than actions.*",
            ):
                dataclasses.replace(traj, acts=traj.acts[:-1])

            with pytest.raises(
                ValueError,
                match=r"infos when present must be present for each action.*",
            ):
                assert traj.infos is not None
                dataclasses.replace(traj, infos=traj.infos[:-1])
            with pytest.raises(
                ValueError,
                match=r"infos when present must be present for each action.*",
            ):
                dataclasses.replace(traj, obs=traj.obs[:-1], acts=traj.acts[:-1])

        _check_1d_shape(
            fn=lambda rews: dataclasses.replace(trajectory_rew, rews=rews),
            length=len(trajectory_rew),
            expected_msg=r"rewards must be 1D array.*",
        )

        with pytest.raises(ValueError, match=r"rewards dtype.* not a float"):
            dataclasses.replace(
                trajectory_rew,
                rews=np.zeros(len(trajectory_rew), dtype=int),
            )

    def test_valid_transitions(
        self,
        transitions_min: types.TransitionsMinimal,
        transitions: types.Transitions,
        transitions_rew: types.TransitionsWithRew,
        length: int,
        n_checks: int = 20,
    ) -> None:
        """Checks initialization, indexing, and slicing sanity."""
        for trans in [transitions_min, transitions, transitions_rew]:
            assert len(trans) == length

            for _ in range(n_checks):
                # Indexing checks, which require at least one element.
                if length != 0:
                    index = np.random.randint(length)
                    assert isinstance(trans[index], dict)
                    _check_transitions_get_item(trans, index)

                # Slicing checks.
                start = np.random.randint(-2, length)
                stop = np.random.randint(0, length + 2)
                step = np.random.randint(-2, 4)
                if step == 0:  # Illegal. Quick fix that biases tests to ordinary step.
                    step = 1
                s = slice(start, stop, step)
                assert type(trans[s]) is type(trans)
                _check_transitions_get_item(trans, s)

    def test_invalid_transitions(
        self,
        transitions_min: types.Transitions,
        transitions: types.Transitions,
        transitions_rew: types.TransitionsWithRew,
        length: int,
    ) -> None:
        """Checks input validation catches space and dtype related errors."""
        if length == 0:
            pytest.skip()

        for trans in [transitions_min, transitions, transitions_rew]:
            with pytest.raises(
                ValueError,
                match=r"obs and acts must have same number of timesteps:.*",
            ):
                dataclasses.replace(trans, acts=trans.acts[:-1])
            with pytest.raises(
                ValueError,
                match=r"obs and infos must have same number of timesteps:.*",
            ):
                dataclasses.replace(trans, infos=[{}] * (length - 1))

        for trans in [transitions, transitions_rew]:
            with pytest.raises(
                ValueError,
                match=r"obs and next_obs must have same shape:.*",
            ):
                dataclasses.replace(trans, next_obs=np.zeros((len(trans), 4, 2)))

            with pytest.raises(
                ValueError,
                match=r"obs and next_obs must have the same dtype:.*",
            ):
                dataclasses.replace(
                    trans,
                    next_obs=np.zeros_like(trans.next_obs, dtype=bool),
                )

            _check_1d_shape(
                fn=lambda bogus_dones: dataclasses.replace(trans, dones=bogus_dones),
                length=len(trans),
                expected_msg=r"dones must be 1D array.*",
            )

            with pytest.raises(ValueError, match=r"dones must be boolean"):
                dataclasses.replace(trans, dones=np.zeros(len(trans), dtype=int))

        _check_1d_shape(
            fn=lambda bogus_rews: dataclasses.replace(trans, rews=bogus_rews),
            length=len(transitions_rew),
            expected_msg=r"rewards must be 1D array.*",
        )

        with pytest.raises(ValueError, match=r"rewards dtype.* not a float"):
            dataclasses.replace(
                transitions_rew,
                rews=np.zeros(len(transitions_rew), dtype=int),
            )


def test_zero_length_fails():
    """Check zero-length trajectory and transitions fail."""
    empty = np.array([])
    with pytest.raises(ValueError, match=r"Degenerate trajectory.*"):
        types.Trajectory(obs=np.array([42]), acts=empty, infos=None, terminal=True)


def test_parse_path():
    if os.name == "nt":  # pragma: no cover
        pytest.skip(
            "Windows uses path.WindowsPath instead when paths are resolved, which"
            "cannot be compared directly to pathlib.Path objects.",
        )
    # absolute paths
    assert util.parse_path("/foo/bar") == pathlib.Path("/foo/bar")
    assert util.parse_path(pathlib.Path("/foo/bar")) == pathlib.Path("/foo/bar")
    assert util.parse_path(b"/foo/bar") == pathlib.Path("/foo/bar")

    # relative paths. implicit conversion to cwd
    assert util.parse_path("foo/bar") == pathlib.Path.cwd() / "foo/bar"
    assert util.parse_path(pathlib.Path("foo/bar")) == pathlib.Path.cwd() / "foo/bar"
    assert util.parse_path(b"foo/bar") == pathlib.Path.cwd() / "foo/bar"

    # relative paths. conversion using custom base directory
    base_dir = pathlib.Path("/foo/bar")
    assert util.parse_path("baz", base_directory=base_dir) == base_dir / "baz"
    assert (
        util.parse_path(pathlib.Path("baz"), base_directory=base_dir)
        == base_dir / "baz"
    )
    assert util.parse_path(b"baz", base_directory=base_dir) == base_dir / "baz"

    # pass a relative path but disallowing relative paths. should raise error.
    with pytest.raises(ValueError, match="Path .* is not absolute"):
        util.parse_path("foo/bar", allow_relative=False)

    # pass a base direectory but disallowing relative paths. should raise error.
    with pytest.raises(
        ValueError,
        match="If `base_directory` is specified, then `allow_relative` must be True.",
    ):
        util.parse_path(
            "foo/bar",
            base_directory=pathlib.Path("/foo/bar"),
            allow_relative=False,
        )

    # Parse optional path. Works the same way but passes None down the line.
    assert util.parse_optional_path(None) is None
    assert util.parse_optional_path("/foo/bar") == util.parse_path("/foo/bar")


def test_dict_obs():
    A = np.random.rand(3, 4)
    B = np.random.rand(3, 7, 1)
    C = np.random.rand(4)

    ab = types.DictObs({"a": A, "b": B})
    abc = types.DictObs({"a": A, "b": B, "c": C})

    # len
    assert len(ab) == 3
    with pytest.raises(RuntimeError):
        len(abc)
    with pytest.raises(RuntimeError):
        len(types.DictObs({}))

    assert abc.dict_len == 3

    # slicing
    np.testing.assert_equal(abc[0].get("a"), A[0])
    np.testing.assert_equal(abc[0].get("c"), np.array(C[0]))
    np.testing.assert_equal(abc[0:2].get("a"), np.array(A[0:2]))
    np.testing.assert_equal(ab[:, 0].get("a"), np.array(A[:, 0]))
    with pytest.raises(IndexError):
        abc[:, 0]

    # iter
    for i, a_row in enumerate(A):
        np.testing.assert_equal(a_row, ab[i].get("a"))
    assert ab[0] == next(iter(ab))

    # eq
    assert abc == types.DictObs({"a": A, "b": B, "c": C})
    assert abc == types.DictObs({"a": np.array(A), "b": np.array(B), "c": np.array(C)})
    assert abc != types.DictObs({"a": A, "c": B, "b": C})  # diff keys
    assert abc != types.DictObs({"a": A, "b": B + 1, "c": C})  # diff values
    assert abc != {"a": A, "b": B + 1, "c": C}  # diff type
    assert abc != ab  # diff keys

    # shape / dtype
    assert abc.shape == {"a": A.shape, "b": B.shape, "c": C.shape}
    assert abc.dtype == {"a": A.dtype, "b": B.dtype, "c": C.dtype}

    # wrap
    assert types.maybe_wrap_in_dictobs({"a": A, "b": B, "c": C}) == abc
    assert abc.unwrap() == {"a": A, "b": B, "c": C}

    # map, stack, concat
    assert abc.map_arrays(lambda arr: arr + 1) == types.DictObs(
        {"a": A + 1, "b": B + 1, "c": C + 1},
    )
    assert types.DictObs.stack(list(iter(ab))) == ab
    np.testing.assert_equal(
        types.DictObs.concatenate([abc, abc]).get("a"),
        np.concatenate([A, A]),
    )

    with pytest.raises(AssertionError):
        types.assert_not_dictobs(abc)

    with pytest.raises(TypeError):
        types.DictObs({"a": "not an array"})  # type: ignore[wrong-arg-types]



"""Tests for `imitation.data.rollout`."""

import functools
from typing import Any, Dict, Mapping, Optional, Sequence, SupportsFloat, Tuple

import gymnasium as gym
import numpy as np
import pytest
from gymnasium.core import WrapperActType, WrapperObsType
from stable_baselines3 import A2C
from stable_baselines3.common import monitor, vec_env
from stable_baselines3.common.env_util import make_atari_env
from stable_baselines3.common.vec_env import VecFrameStack
from stable_baselines3.ppo import MlpPolicy

from imitation.data import rollout, types, wrappers
from imitation.policies import serialize
from imitation.policies.base import RandomPolicy


class TerminalSentinelEnv(gym.Env):
    """Environment with observation 0 when alive and 1 at terminal state."""

    def __init__(self, max_acts: int):
        """Builds `TerminalSentinelLength` with episode length `max_acts`."""
        self.max_acts = max_acts
        self.current_step = 0
        self.action_space = gym.spaces.Discrete(1)
        self.observation_space = gym.spaces.Box(np.array([0]), np.array([1]))

    def reset(self, seed=None):
        self.current_step = 0
        return np.array([0]), {}

    def step(self, action):
        self.current_step += 1
        done = self.current_step >= self.max_acts
        observation = np.array([1 if done else 0])
        rew = 0.0
        return observation, rew, done, False, {}


def _sample_fixed_length_trajectories(
    episode_lengths: Sequence[int],
    min_episodes: int,
    rng: np.random.Generator,
    policy_type: str = "policy",
    **kwargs,
) -> Sequence[types.Trajectory]:
    venv = vec_env.DummyVecEnv(
        [functools.partial(TerminalSentinelEnv, length) for length in episode_lengths],
    )
    policy: rollout.AnyPolicy
    if policy_type == "policy":
        policy = RandomPolicy(venv.observation_space, venv.action_space)
    elif policy_type == "callable":
        random_policy = RandomPolicy(venv.observation_space, venv.action_space)

        # Simple way to get a valid callable: just use a policies .predict() method
        # (still tests another code path inside generate_trajectories)
        def policy(x, state, mask):
            return random_policy.predict(x, state, mask)

    elif policy_type == "random":
        policy = None
    else:  # pragma: no cover
        raise ValueError(f"Unknown policy_type '{policy_type}'")
    sample_until = rollout.make_min_episodes(min_episodes)
    trajectories = rollout.generate_trajectories(
        policy,
        venv,
        sample_until=sample_until,
        rng=rng,
        **kwargs,
    )
    return trajectories


@pytest.mark.parametrize(
    "policy_type",
    ["policy", "callable", "random"],
)
def test_complete_trajectories(policy_type, rng) -> None:
    """Checks trajectories include the terminal observation.

    This is hidden by default by VecEnv's auto-reset; we add it back in using
    `rollout.RolloutInfoWrapper`.

    Args:
        policy_type: Kind of policy to use when generating trajectories.
        rng: Random state to use.
    """
    min_episodes = 13
    max_acts = 5
    num_envs = 4
    trajectories = _sample_fixed_length_trajectories(
        [max_acts] * num_envs,
        min_episodes,
        policy_type=policy_type,
        rng=rng,
    )
    assert len(trajectories) >= min_episodes
    expected_obs = np.array([[0]] * max_acts + [[1]])
    for trajectory in trajectories:
        obs = trajectory.obs
        acts = trajectory.acts
        assert len(obs) == len(acts) + 1
        assert np.all(obs == expected_obs)


@pytest.mark.parametrize(
    "episode_lengths,min_episodes,expected_counts",
    [
        # Do we keep on sampling from the 1st (len 3) environment that remains 'alive'?
        ([3, 5], 2, {3: 2, 5: 1}),
        # Do we keep on sampling from the 2nd (len 7) environment that remains 'alive'?
        ([3, 7], 2, {3: 2, 7: 1}),
        # Similar, but more extreme case with num environments > num episodes
        ([3, 3, 3, 7], 2, {3: 3, 7: 1}),
        # Do we stop sampling at 2 episodes if we get two equal-length episodes?
        ([3, 3], 2, {3: 2}),
        ([5, 5], 2, {5: 2}),
        ([7, 7], 2, {7: 2}),
    ],
)
def test_unbiased_trajectories(
    episode_lengths: Sequence[int],
    min_episodes: int,
    expected_counts: Mapping[int, int],
    rng,
) -> None:
    """Checks trajectories are sampled without bias towards shorter episodes.

    Specifically, we create a VecEnv consisting of environments with fixed-length
    `episode_lengths`. This is unrealistic and breaks the i.i.d. assumption, but lets
    us test things deterministically.

    If we hit `min_episodes` exactly and all environments are done at the same time,
    we should stop and not sample any more trajectories. Otherwise, we should keep
    sampling from any in-flight environments, but not add trajectories from any other
    environments.

    The different test cases check each of these cases.

    Args:
        episode_lengths: The length of the episodes in each environment.
        min_episodes: The minimum number of episodes to sample.
        expected_counts: Mapping from episode length to expected number of episodes
            of that length (omit if 0 episodes of that length expected).
        rng: Random state to use.
    """
    trajectories = _sample_fixed_length_trajectories(
        episode_lengths,
        min_episodes,
        rng,
    )
    assert len(trajectories) == sum(expected_counts.values())
    traj_lens = np.array([len(traj) for traj in trajectories])
    for length, count in expected_counts.items():
        assert np.sum(traj_lens == length) == count


def test_seed_trajectories():
    """Check trajectory order deterministic given seed and that seed is not no-op.

    Note in general environments and policies are stochastic, so the trajectory
    order *will* differ unless environment/policy seeds are also set.

    However, `TerminalSentinelEnv` is fixed-length deterministic, so there are no
    such confounders in this test.
    """
    rng_a1 = np.random.default_rng(0)
    rng_a2 = np.random.default_rng(0)
    rng_b = np.random.default_rng(1)
    traj_a1 = _sample_fixed_length_trajectories([3, 5], 2, rng=rng_a1)
    traj_a2 = _sample_fixed_length_trajectories([3, 5], 2, rng=rng_a2)
    traj_b = _sample_fixed_length_trajectories([3, 5], 2, rng=rng_b)
    assert [len(traj) for traj in traj_a1] == [len(traj) for traj in traj_a2]
    assert [len(traj) for traj in traj_a1] != [len(traj) for traj in traj_b]


class ObsRewHalveWrapper(gym.Wrapper):
    """Simple wrapper that scales every reward and observation feature by 0.5."""

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> Tuple[WrapperObsType, Dict[str, Any]]:
        obs, info = super().reset(seed=seed, options=options)
        return obs / 2.0, info

    def step(
        self,
        action: WrapperActType,
    ) -> Tuple[WrapperObsType, SupportsFloat, bool, bool, Dict[str, Any]]:
        obs, rew, terminated, truncated, info = super().step(action)
        return obs / 2, float(rew) / 2, terminated, truncated, info


def test_rollout_stats(rng):
    """Applying `ObsRewIncrementWrapper` halves the reward mean.

    `rollout_stats` should reflect this.

    Args:
        rng: Random state to use (with fixed seed).
    """
    env = gym.make("CartPole-v1")
    env = monitor.Monitor(env, None)
    env = ObsRewHalveWrapper(env)
    venv = vec_env.DummyVecEnv([lambda: env])

    policy = serialize.load_policy("zero", venv)
    trajs = rollout.generate_trajectories(
        policy,
        venv,
        rollout.make_min_episodes(10),
        rng=rng,
    )
    s = rollout.rollout_stats(trajs)

    np.testing.assert_allclose(s["return_mean"], s["monitor_return_mean"] / 2)
    np.testing.assert_allclose(s["return_std"], s["monitor_return_std"] / 2)
    np.testing.assert_allclose(s["return_min"], s["monitor_return_min"] / 2)
    np.testing.assert_allclose(s["return_max"], s["monitor_return_max"] / 2)


def test_unwrap_traj(rng):
    """Check that unwrap_traj reverses `ObsRewIncrementWrapper`.

    Also check that unwrapping twice is a no-op.

    Args:
        rng: Random state to use (with fixed seed).
    """
    env = gym.make("CartPole-v1")
    env = wrappers.RolloutInfoWrapper(env)
    env = ObsRewHalveWrapper(env)
    venv = vec_env.DummyVecEnv([lambda: env])

    policy = serialize.load_policy("zero", venv)
    trajs = rollout.generate_trajectories(
        policy,
        venv,
        rollout.make_min_episodes(10),
        rng=rng,
    )
    trajs_unwrapped = [rollout.unwrap_traj(t) for t in trajs]
    trajs_unwrapped_twice = [rollout.unwrap_traj(t) for t in trajs_unwrapped]

    for t, t_unwrapped in zip(trajs, trajs_unwrapped):
        np.testing.assert_allclose(t.acts, t_unwrapped.acts)
        np.testing.assert_allclose(t.obs, t_unwrapped.obs / 2)
        np.testing.assert_allclose(t.rews, t_unwrapped.rews / 2)

    for t1, t2 in zip(trajs_unwrapped, trajs_unwrapped_twice):
        np.testing.assert_equal(t1.acts, t2.acts)
        np.testing.assert_equal(t1.obs, t2.obs)
        np.testing.assert_equal(t1.rews, t2.rews)


def test_unwrap_traj_raises_no_infos():
    """Check that unwrap_traj raises ValueError if no infos in trajectory."""
    with pytest.raises(ValueError, match="Trajectory must have infos to unwrap"):
        acts = np.array([0])
        obs = np.array([0, 0])
        rews = np.array([0.0])
        rollout.unwrap_traj(
            types.TrajectoryWithRew(
                acts=acts,
                obs=obs,
                terminal=False,
                rews=rews,
                infos=None,
            ),
        )


def test_make_sample_until_errors():
    with pytest.raises(ValueError, match="At least one.*"):
        rollout.make_sample_until(min_timesteps=None, min_episodes=None)

    episodes_positive = pytest.raises(ValueError, match="min_episodes.*positive")
    with episodes_positive:
        rollout.make_sample_until(min_timesteps=None, min_episodes=0)
    with episodes_positive:
        rollout.make_sample_until(min_timesteps=10, min_episodes=-34)

    timesteps_positive = pytest.raises(ValueError, match="min_timesteps.*positive")
    with timesteps_positive:
        rollout.make_sample_until(min_timesteps=-3, min_episodes=None)
    with timesteps_positive:
        rollout.make_sample_until(min_timesteps=0, min_episodes=None)


@pytest.mark.parametrize("gamma", [0, 0.9, 1])
def test_compute_returns(gamma):
    rng = np.random.default_rng(seed=0)
    N = 100
    rewards = rng.random(N)
    discounts = np.power(gamma, np.arange(N))
    returns = np.sum(discounts * rewards)
    # small numerical errors will occur because compute_returns
    # uses a somewhat different method based on evaluating
    # polynomials
    assert abs(rollout.discounted_sum(rewards, gamma) - returns) < 1e-8


def test_generate_trajectories_type_error(rng):
    venv = vec_env.DummyVecEnv([functools.partial(TerminalSentinelEnv, 1)])
    sample_until = rollout.make_min_episodes(1)
    with pytest.raises(TypeError, match="Policy must be.*got <class 'str'> instead"):
        rollout.generate_trajectories(
            "strings_are_not_valid_policies",  # type: ignore[arg-type]
            venv,
            rng=rng,
            sample_until=sample_until,
        )


def test_generate_trajectories_value_error(rng):
    venv = vec_env.DummyVecEnv([functools.partial(TerminalSentinelEnv, 1)])
    sample_until = rollout.make_min_episodes(1)

    with pytest.raises(ValueError, match="Cannot set deterministic.*is ignored."):
        rollout.generate_trajectories(
            lambda obs: np.zeros(len(obs), dtype=int),
            venv,
            sample_until=sample_until,
            rng=rng,
            deterministic_policy=True,
        )


def test_rollout_verbose_error_for_image_environments(rng):
    seed = 0
    env = make_atari_env("BeamRiderNoFrameskip-v4", n_envs=1, seed=seed)
    env = VecFrameStack(env, n_stack=4)
    expert = A2C(policy=MlpPolicy, env=env, seed=seed)
    expert.learn(1)

    rng = np.random.default_rng(seed)
    with pytest.raises(
        ValueError,
        match=r".*expert\.get_env().*",
    ):
        rollout.rollout(
            expert,
            # Note that this should be expert.get_env(), which rearranges
            # the channel dimension to the first position.
            env,
            rollout.make_sample_until(min_timesteps=None, min_episodes=2),
            rng=rng,
        )


def test_rollout_normal_error_for_other_shape_mismatch(rng):
    seed = 0
    env = make_atari_env("BeamRiderNoFrameskip-v4", n_envs=1, seed=seed)
    env = VecFrameStack(env, n_stack=4)
    expert = A2C(policy=MlpPolicy, env=env, seed=seed)
    expert.learn(1)

    rng = np.random.default_rng(seed)
    unrelated_env = vec_env.DummyVecEnv(
        [functools.partial(TerminalSentinelEnv, 5)],
    )
    other_image_env = make_atari_env("Atlantis-v0", n_envs=1, seed=seed)
    print(other_image_env.observation_space.shape)
    for bad_env in [other_image_env, unrelated_env]:
        with pytest.raises(
            ValueError,
            match=r"Observation spaces do not match.*",
        ):
            rollout.rollout(
                expert,
                bad_env,
                rollout.make_sample_until(min_timesteps=None, min_episodes=2),
                rng=rng,
            )


class DictObsWrapper(gym.ObservationWrapper):
    """Simple wrapper that turns the observation into a dictionary.

    The observation is duplicated, with "b" rescaled.
    """

    def __init__(self, env: gym.Env):
        """Builds DictObsWrapper.

        Args:
            env: The wrapped Env.
        """
        super().__init__(env)
        self.observation_space = gym.spaces.Dict(
            {"a": env.observation_space, "b": env.observation_space},
        )

    def observation(self, observation):
        return {"a": observation, "b": observation / 2}


def test_dictionary_observations(rng):
    """Test we can generate a rollout for a dict-type observation environment.

    Args:
        rng: Random state to use (with fixed seed).
    """
    env = gym.make("CartPole-v1")
    env = monitor.Monitor(env, None)
    env = DictObsWrapper(env)
    venv = vec_env.DummyVecEnv([lambda: env])

    policy = serialize.load_policy("zero", venv)
    trajs = rollout.generate_trajectories(
        policy,
        venv,
        rollout.make_min_episodes(10),
        rng=rng,
    )
    for traj in trajs:
        assert isinstance(traj.obs, types.DictObs)
        for obs in traj.obs:
            assert venv.observation_space.contains(dict(obs.items()))
        np.testing.assert_allclose(traj.obs.get("a") / 2, traj.obs.get("b"))



"""Interactive policies that query the user for actions."""

import abc
import collections
from typing import Dict, Optional, Union

import gymnasium as gym
import matplotlib.pyplot as plt
import numpy as np
from shimmy import atari_env
from stable_baselines3.common import vec_env

import imitation.policies.base as base_policies
from imitation.util import util


class DiscreteInteractivePolicy(base_policies.NonTrainablePolicy, abc.ABC):
    """Abstract class for interactive policies with discrete actions.

    For each query, the observation is rendered and then the action is provided
    as a keyboard input.
    """

    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.Space,
        action_keys_names: collections.OrderedDict,
        clear_screen_on_query: bool = True,
    ):
        """Builds DiscreteInteractivePolicy.

        Args:
            observation_space: Observation space.
            action_space: Action space.
            action_keys_names: `OrderedDict` containing pairs (key, name) for every
                action, where key will be used in the console interface, and name
                is a semantic action name. The index of the pair in the dictionary
                will be used as the discrete, integer action.
            clear_screen_on_query: If `True`, console will be cleared on every query.
        """
        super().__init__(
            observation_space=observation_space,
            action_space=action_space,
        )

        assert isinstance(action_space, gym.spaces.Discrete)
        assert (
            len(action_keys_names)
            == len(set(action_keys_names.values()))
            == action_space.n
        )

        self.action_keys_names = action_keys_names
        self.action_key_to_index = {
            k: i for i, k in enumerate(action_keys_names.keys())
        }
        self.clear_screen_on_query = clear_screen_on_query

    def _choose_action(
        self,
        obs: Union[np.ndarray, Dict[str, np.ndarray]],
    ) -> np.ndarray:
        if self.clear_screen_on_query:
            util.clear_screen()

        if isinstance(obs, dict):
            raise ValueError("Dictionary observations are not supported here")

        context = self._render(obs)
        key = self._get_input_key()
        self._clean_up(context)

        return np.array([self.action_key_to_index[key]])

    def _get_input_key(self) -> str:
        """Obtains input key for action selection."""
        print(
            "Please select an action. Possible choices in [ACTION_NAME:KEY] format:",
            ", ".join([f"{n}:{k}" for k, n in self.action_keys_names.items()]),
        )

        key = input("Your choice (enter key):")
        while key not in self.action_keys_names.keys():
            key = input("Invalid key, please try again! Your choice (enter key):")

        return key

    @abc.abstractmethod
    def _render(self, obs: np.ndarray) -> Optional[object]:
        """Renders an observation, optionally returns a context for later cleanup."""

    def _clean_up(self, context: object) -> None:
        """Cleans up after the input has been captured, e.g. stops showing the image."""


class ImageObsDiscreteInteractivePolicy(DiscreteInteractivePolicy):
    """DiscreteInteractivePolicy that renders image observations."""

    def _render(self, obs: np.ndarray) -> plt.Figure:
        img = self._prepare_obs_image(obs)

        fig, ax = plt.subplots()
        ax.imshow(img, cmap="gray", vmin=0, vmax=255)  # cmap is ignored for RGB images.
        ax.axis("off")
        fig.show()

        return fig

    def _clean_up(self, context: plt.Figure) -> None:
        plt.close(context)

    def _prepare_obs_image(self, obs: np.ndarray) -> np.ndarray:
        """Applies any required observation processing to get an image to show."""
        return obs


ATARI_ACTION_NAMES_TO_KEYS = {
    "NOOP": "1",
    "FIRE": "2",
    "UP": "w",
    "RIGHT": "d",
    "LEFT": "a",
    "DOWN": "x",
    "UPRIGHT": "e",
    "UPLEFT": "q",
    "DOWNRIGHT": "c",
    "DOWNLEFT": "z",
    "UPFIRE": "t",
    "RIGHTFIRE": "h",
    "LEFTFIRE": "f",
    "DOWNFIRE": "b",
    "UPRIGHTFIRE": "y",
    "UPLEFTFIRE": "r",
    "DOWNRIGHTFIRE": "n",
    "DOWNLEFTFIRE": "v",
}


class AtariInteractivePolicy(ImageObsDiscreteInteractivePolicy):
    """Interactive policy for Atari environments."""

    def __init__(self, env: Union[atari_env.AtariEnv, vec_env.VecEnv], *args, **kwargs):
        """Builds AtariInteractivePolicy."""
        action_names = (
            env.get_action_meanings()
            if isinstance(env, atari_env.AtariEnv)
            else env.env_method("get_action_meanings", indices=[0])[0]
        )
        action_keys_names = collections.OrderedDict(
            [(ATARI_ACTION_NAMES_TO_KEYS[name], name) for name in action_names],
        )
        super().__init__(
            env.observation_space,
            env.action_space,
            action_keys_names,
            *args,
            **kwargs,
        )



"""Test `imitation.algorithms.tabular_irl` and tabular environments."""

from typing import Any, Mapping, cast

import gymnasium as gym
import numpy as np
import pytest
import torch as th
from seals import base_envs
from seals.diagnostics import random_trans
from stable_baselines3.common import vec_env

from imitation.algorithms import base
from imitation.algorithms.mce_irl import (
    MCEIRL,
    TabularPolicy,
    mce_occupancy_measures,
    mce_partition_fh,
)
from imitation.data import rollout
from imitation.rewards import reward_nets
from imitation.util.util import tensor_iter_norm


def rollouts(env, n=10, seed=None):
    rv = []
    for _ in range(n):
        done = False
        # if a seed is given, then we use the same seed each time (should
        # give same trajectory each time)
        obs = env.reset(seed=seed)
        if seed is not None:
            env.action_space.seed(seed)

        traj = [obs]
        while not done:
            act = env.action_space.sample()
            obs, rew, terminated, truncated, _ = env.step(act)
            done = terminated or truncated
            traj.append((obs, rew))
        rv.append(traj)
    return rv


@pytest.fixture
def random_mdp():
    return random_trans.RandomTransitionEnv(
        n_states=5,
        n_actions=3,
        branch_factor=2,
        horizon=10,
        random_obs=False,
        obs_dim=None,
        generator_seed=42,
    )


def make_reward_net(env: gym.Env) -> reward_nets.BasicRewardNet:
    """Makes linear reward model."""
    return reward_nets.BasicRewardNet(
        env.observation_space,
        env.action_space,
        use_action=False,
        use_next_state=False,
        use_done=False,
        hid_sizes=[],
    )


def test_random_mdp():
    for i in range(3):
        n_states = 4 * (i + 3)
        n_actions = i + 2
        branch_factor = i + 1
        if branch_factor == 1:
            # make sure we have enough actions to get reasonable trajectories
            n_actions = min(n_states, max(n_actions, 4))
        horizon = 5 * (i + 1)
        random_obs = (i % 2) == 0
        obs_dim = (i * 3 + 4) ** 2 + i
        mdp = random_trans.RandomTransitionEnv(
            n_states=n_states,
            n_actions=n_actions,
            branch_factor=branch_factor,
            horizon=horizon,
            random_obs=random_obs,
            obs_dim=obs_dim if random_obs else None,
            generator_seed=i,
        )

        # sanity checks on sizes of things
        assert mdp.transition_matrix.shape == (n_states, n_actions, n_states)
        assert np.allclose(1, np.sum(mdp.transition_matrix, axis=-1))
        assert np.all(mdp.transition_matrix >= 0)
        assert (
            mdp.observation_matrix.shape[0] == n_states
            and mdp.observation_matrix.ndim == 2
        )
        assert mdp.reward_matrix.shape == (n_states,)
        assert mdp.horizon == horizon
        assert np.all(mdp.initial_state_dist >= 0)
        assert np.allclose(1, np.sum(mdp.initial_state_dist))
        assert np.sum(mdp.initial_state_dist > 0) == branch_factor

        # make sure trajectories aren't all the same if we don't specify same
        # seed each time
        trajectories = rollouts(mdp, 100)
        assert len(set(map(str, trajectories))) > 1

        trajectories = rollouts(mdp, 100, seed=42)
        # make sure trajectories ARE all the same if we do specify the same
        # seed each time
        assert len(set(map(str, trajectories))) == 1


def test_infinite_horizon_error(random_mdp, rng):
    random_mdp.horizon = None
    check_raises = pytest.raises(ValueError, match="Only finite-horizon.*")
    with check_raises:
        mce_partition_fh(random_mdp)
    with check_raises:
        mce_occupancy_measures(random_mdp)
    reward_net = make_reward_net(random_mdp)
    with check_raises:
        MCEIRL(None, random_mdp, reward_net, rng)


FEW_DISCOUNT_RATES = [0.0, 0.99, 1.0]
DISCOUNT_RATES = FEW_DISCOUNT_RATES + [0.5, 0.9]


@pytest.mark.parametrize("discount", DISCOUNT_RATES)
def test_policy_om_random_mdp(discount: float):
    """Test that optimal policy occupancy measure ("om") for a random MDP is sane."""
    mdp = gym.make("seals/Random-v0")
    V, Q, pi = mce_partition_fh(
        cast(base_envs.TabularModelPOMDP, mdp),
        discount=discount,
    )
    assert np.all(np.isfinite(V))
    assert np.all(np.isfinite(Q))
    assert np.all(np.isfinite(pi))
    # Check it is a probability distribution along the last axis
    assert np.all(pi >= 0)
    assert np.allclose(np.sum(pi, axis=-1), 1)

    horizon = cast(base_envs.TabularModelPOMDP, mdp).horizon
    Dt, D = mce_occupancy_measures(
        cast(base_envs.TabularModelPOMDP, mdp),
        pi=pi,
        discount=discount,
    )
    assert isinstance(horizon, int)
    assert len(Dt) == horizon + 1
    assert np.all(np.isfinite(D))
    assert np.any(D > 0)
    # expected number of state visits (over all states) should be equal to the
    # horizon
    if discount == 1.0:
        expected_sum = horizon + 1.0
    else:
        expected_sum = (1 - discount ** (horizon + 1)) / (1 - discount)
    assert np.allclose(np.sum(D), expected_sum)


class ReasonablePOMDP(base_envs.TabularModelPOMDP):
    """A tabular MDP with sensible parameters."""

    def __init__(self):
        """Initialize a ReasonablePOMDP."""
        observation_matrix = np.array(
            [
                [3, -5, -1, -1, -4, 5, 3, 0],
                # state 1 (top)
                [4, -4, 2, 2, -4, -1, -2, -2],
                # state 2 (bottom, equiv to top)
                [3, -1, 5, -1, 0, 2, -5, 2],
                # state 3 (middle, very low reward and so dominated by others)
                [-5, -1, 4, 1, 4, 1, 5, 3],
                # state 4 (final, all self loops, good reward)
                [2, -5, 1, -5, 1, 4, 4, -3],
            ],
        )
        transition_matrix = np.array(
            [
                # transitions out of state 0
                [
                    # action 0: goes to state 1 (sometimes 2)
                    [0, 0.9, 0.1, 0, 0],
                    # action 1: goes to state 3 deterministically
                    [0, 0, 0, 1, 0],
                    # action 2: goes to state 2 (sometimes 2)
                    [0, 0.1, 0.9, 0, 0],
                ],
                # transitions out of state 1
                [
                    # action 0: goes to state 3 or 4 (sub-optimal)
                    [0, 0, 0, 0.05, 0.95],
                    # action 1: goes to state 3 (bad)
                    [0, 0, 0, 1, 0],
                    # action 2: goes to state 4 (good!)
                    [0, 0, 0, 0, 1],
                ],
                # transitions out of state 2 (basically the same)
                [
                    # action 0: goes to state 3 or 4 (sub-optimal)
                    [0, 0, 0, 0.05, 0.95],
                    # action 1: goes to state 3 (bad)
                    [0, 0, 0, 1, 0],
                    # action 2: goes to state 4 (good!)
                    [0, 0, 0, 0, 1],
                ],
                # transitions out of state 3 (all go to state 4)
                [
                    # action 0
                    [0, 0, 0, 0, 1],
                    # action 1
                    [0, 0, 0, 0, 1],
                    # action 2
                    [0, 0, 0, 0, 1],
                ],
                # transitions out of state 4 (all go back to state 0)
                [
                    # action 0
                    [1, 0, 0, 0, 0],
                    # action 1
                    [1, 0, 0, 0, 0],
                    # action 2
                    [1, 0, 0, 0, 0],
                ],
            ],
        )
        reward_matrix = np.array(
            [
                # state 0 (okay reward, but we can't go back so it doesn't matter)
                1,
                # states 1 & 2 have same (okay) reward
                2,
                2,
                # state 3 has very negative reward (so avoid it!)
                -20,
                # state 4 has pretty good reward (good enough that we should move out
                # of 1 & 2)
                3,
            ],
        )
        # always start in s0 or s4
        initial_state_dist = np.array([0.5, 0.0, 0.0, 0.0, 0.5])
        horizon = 20
        super().__init__(
            observation_matrix=observation_matrix,
            transition_matrix=transition_matrix,
            reward_matrix=reward_matrix,
            initial_state_dist=initial_state_dist,
            horizon=horizon,
        )


@pytest.mark.parametrize("discount", DISCOUNT_RATES)
def test_policy_om_reasonable_pomdp(discount: float):
    # MDP described above
    pomdp = ReasonablePOMDP()
    # get policy etc. for our MDP
    V, Q, pi = mce_partition_fh(pomdp, discount=discount)
    Dt, D = mce_occupancy_measures(pomdp, pi=pi, discount=discount)
    assert np.all(np.isfinite(V))
    assert np.all(np.isfinite(Q))
    assert np.all(np.isfinite(pi))
    assert np.all(np.isfinite(Dt))
    assert np.all(np.isfinite(D))
    # check that actions 0 & 2 (which go to states 1 & 2) are roughly equal
    assert np.allclose(pi[:19, 0, 0], pi[:19, 0, 2])
    # also check that they're by far preferred to action 1 (that goes to state
    # 3, which has poor reward)
    if discount > 0:
        assert np.all(pi[:19, 0, 0] > 2 * pi[:19, 0, 1])
    # make sure that states 3 & 4 have roughly uniform policies
    pi_34 = pi[:5, 3:5]
    assert np.allclose(pi_34, np.ones_like(pi_34) / 3.0)
    # check that states 1 & 2 have similar policies to each other
    assert np.allclose(pi[:19, 1, :], pi[:19, 2, :])
    # check that in state 1, action 2 (which goes to state 4 with certainty) is
    # better than action 0 (which only gets there with some probability), and
    # that both are better than action 1 (which always goes to the bad state).
    if discount > 0:
        assert np.all(pi[:19, 1, 2] > pi[:19, 1, 0])
        assert np.all(pi[:19, 1, 0] > pi[:19, 1, 1])
    # check that Dt[0] matches our initial state dist
    assert np.allclose(Dt[0], pomdp.initial_state_dist)


def test_tabular_policy(rng):
    """Tests tabular policy prediction, especially timestep calculation and masking."""
    state_space = gym.spaces.Discrete(2)
    action_space = gym.spaces.Discrete(2)
    pi = np.stack(
        [np.eye(2), 1 - np.eye(2)],
    )
    tabular = TabularPolicy(
        state_space=state_space,
        action_space=action_space,
        pi=pi,
        rng=rng,
    )

    states = np.array([0, 1, 1, 0, 1])
    actions, timesteps = tabular.predict(states)
    np.testing.assert_array_equal(states, actions)
    np.testing.assert_equal(timesteps[0], 1)

    mask = np.zeros((5,), dtype=bool)
    actions, timesteps = tabular.predict(states, timesteps, mask)
    np.testing.assert_array_equal(1 - states, actions)
    np.testing.assert_equal(timesteps[0], 2)

    mask = np.ones((5,), dtype=bool)
    actions, timesteps = tabular.predict(states, timesteps, mask)
    np.testing.assert_array_equal(states, actions)
    np.testing.assert_equal(timesteps[0], 1)

    mask = (1 - states).astype(bool)
    actions, timesteps = tabular.predict(states, timesteps, mask)
    np.testing.assert_array_equal(np.zeros((5,)), actions)
    np.testing.assert_equal(timesteps[0], 2 - mask.astype(int))


def test_tabular_policy_rollouts(rng):
    """Tests that rolling out a tabular policy that varies at each timestep works."""
    state_space = gym.spaces.Discrete(5)
    action_space = gym.spaces.Discrete(3)
    mdp = ReasonablePOMDP()
    state_env = base_envs.ExposePOMDPStateWrapper(mdp)
    state_venv = vec_env.DummyVecEnv([lambda: state_env])

    # alternate actions every step
    subpolicy = np.stack([np.eye(action_space.n)] * state_space.n, axis=1)

    # repeat 7 times for a total of 21 (greater than 20)
    pi = np.repeat(
        subpolicy,
        ((mdp.horizon + action_space.n - 1) // action_space.n),
        axis=0,
    )

    tabular = TabularPolicy(
        state_space=state_space,
        action_space=action_space,
        pi=pi,
        rng=rng,
    )

    trajs = rollout.generate_trajectories(
        tabular,
        state_venv,
        sample_until=rollout.make_min_episodes(1),
        rng=rng,
    )

    # pi[t,s,a] is the same for every state, so drop that dimension
    exposed_actions_onehot = pi[:, 0, :]
    exposed_actions = exposed_actions_onehot.nonzero()[1]

    # check that the trajectory chooses the same actions as the policy
    assert (trajs[0].acts == exposed_actions[: len(trajs[0].acts)]).all()


def test_tabular_policy_randomness(rng):
    state_space = gym.spaces.Discrete(2)
    action_space = gym.spaces.Discrete(2)
    pi = np.array(
        [
            [
                [0.5, 0.5],
                [0.9, 0.1],
            ],
        ],
    )
    tabular = TabularPolicy(
        state_space=state_space,
        action_space=action_space,
        pi=pi,
        rng=rng,
    )

    actions, _ = tabular.predict(np.zeros((1000,), dtype=int))
    assert 0.45 <= np.mean(actions) <= 0.55
    ones_obs = np.ones((1000,), dtype=int)
    actions, _ = tabular.predict(ones_obs)
    assert 0.05 <= np.mean(actions) <= 0.15
    actions, _ = tabular.predict(ones_obs, deterministic=True)
    np.testing.assert_equal(actions, 0)


def test_mce_irl_demo_formats(rng, random_mdp):
    state_env = base_envs.ExposePOMDPStateWrapper(random_mdp)
    state_venv = vec_env.DummyVecEnv([lambda: state_env])
    trajs = rollout.generate_trajectories(
        policy=None,
        venv=state_venv,
        sample_until=rollout.make_min_timesteps(100),
        rng=rng,
    )
    demonstrations = {
        "trajs": trajs,
        "trans": rollout.flatten_trajectories(trajs),
        "data_loader": base.make_data_loader(
            trajs,
            batch_size=32,
            data_loader_kwargs=dict(drop_last=False),
        ),
    }

    final_counts = {}
    for kind, demo in demonstrations.items():
        with th.random.fork_rng():
            th.random.manual_seed(715298)
            # create reward network so we can be sure it's seeded identically
            reward_net = make_reward_net(random_mdp)
            mce_irl = MCEIRL(
                demo,
                random_mdp,
                reward_net,
                linf_eps=1e-3,
                rng=rng,
            )
            assert np.allclose(mce_irl.demo_state_om.sum(), random_mdp.horizon + 1)
            final_counts[kind] = mce_irl.train(max_iter=5)

            # make sure weights have non-insane norm
            assert tensor_iter_norm(mce_irl.reward_net.parameters()) < 1000

    for k, cts in final_counts.items():
        assert np.allclose(cts, final_counts["trajs"], atol=1e-3, rtol=1e-3), k


@pytest.mark.expensive
@pytest.mark.parametrize(
    "model_kwargs",
    [dict(hid_sizes=[]), dict(hid_sizes=[32, 32])],
)
@pytest.mark.parametrize("discount", FEW_DISCOUNT_RATES)
def test_mce_irl_reasonable_mdp(
    model_kwargs: Mapping[str, Any],
    discount: float,
    rng,
):
    with th.random.fork_rng():
        th.random.manual_seed(715298)

        # test MCE IRL on the MDP
        mdp = ReasonablePOMDP()
        mdp.reset(seed=715298)

        # demo occupancy measure
        V, Q, pi = mce_partition_fh(mdp, discount=discount)
        Dt, D = mce_occupancy_measures(mdp, pi=pi, discount=discount)

        reward_net = reward_nets.BasicRewardNet(
            mdp.observation_space,
            mdp.action_space,
            use_action=False,
            use_next_state=False,
            use_done=False,
            **model_kwargs,
        )
        mce_irl = MCEIRL(
            D,
            mdp,
            reward_net,
            linf_eps=1e-3,
            discount=discount,
            rng=rng,
        )
        final_counts = mce_irl.train()

        assert np.allclose(final_counts, D, atol=1e-3, rtol=1e-3)
        # make sure weights have non-insane norm
        assert tensor_iter_norm(reward_net.parameters()) < 1000

        state_env = base_envs.ExposePOMDPStateWrapper(mdp)
        state_venv = vec_env.DummyVecEnv([lambda: state_env])
        trajs = rollout.generate_trajectories(
            mce_irl.policy,
            state_venv,
            sample_until=rollout.make_min_episodes(5),
            rng=rng,
        )
        stats = rollout.rollout_stats(trajs)
        if discount > 0.0:  # skip check when discount==0.0 (random policy)
            eps = 1e-6  # avoid test failing due to rounding error
            assert mdp.horizon is not None
            assert stats["return_mean"] >= (mdp.horizon - 1) * 2 * 0.8 - eps



"""Tests for config files in imitation/scripts/config/tuned_hps/ folder."""

import pytest

from imitation.scripts import train_adversarial, train_imitation, tuning

ALGORITHMS = ["bc", "dagger", "airl", "gail"]
ENVIRONMENTS = [
    "seals_walker",
    "seals_ant",
    "seals_half_cheetah",
    "seals_hopper",
    "seals_swimmer",
]


@pytest.mark.parametrize("environment", ENVIRONMENTS)
@pytest.mark.parametrize("algorithm", ALGORITHMS)
def test_benchmarks_print_config_succeeds(algorithm: str, environment: str):
    # We test the configs using the print_config command,
    # because running the configs requires MuJoCo.
    # Requiring MuJoCo to run the tests adds too much complexity.

    if algorithm in ("bc", "dagger"):
        experiment = train_imitation.train_imitation_ex
    elif algorithm in ("airl", "gail"):
        experiment = train_adversarial.train_adversarial_ex
    else:
        raise ValueError(f"Unknown algorithm: {algorithm}")  # pragma: no cover

    config_name = f"{algorithm}_{environment}"
    run = experiment.run(command_name="print_config", named_configs=[config_name])
    assert run.status == "COMPLETED"


@pytest.mark.parametrize("environment", ENVIRONMENTS)
@pytest.mark.parametrize("algorithm", ALGORITHMS)
def test_tuning_print_config_succeeds(algorithm: str, environment: str):
    # We test the configs using the print_config command,
    # because running the configs requires MuJoCo.
    # Requiring MuJoCo to run the tests adds too much complexity.
    experiment = tuning.tuning_ex
    run = experiment.run(
        command_name="print_config",
        named_configs=[algorithm],
        config_updates=dict(
            parallel_run_config=dict(
                base_named_configs=[f"{algorithm}_{environment}"],
            ),
        ),
    )
    assert run.status == "COMPLETED"



"""Command-line scripts."""

# Add our custom environments to Gym registry.

try:
    # pytype: disable=import-error
    import seals  # noqa: F401

    # pytype: enable=import-error
except ImportError:
    pass



"""Uses RL to train a policy from scratch, saving rollouts and policy.

This can be used:
    1. To train a policy on a ground-truth reward function, as a source of
       synthetic "expert" demonstrations to train IRL or imitation learning
       algorithms.
    2. To train a policy on a learned reward function, to solve a task or
       as a way of evaluating the quality of the learned reward function.
"""

import logging
import pathlib
import warnings
from typing import Any, Mapping, Optional

import numpy as np
from sacred.observers import FileStorageObserver
from stable_baselines3.common import callbacks
from stable_baselines3.common.vec_env import VecNormalize

import imitation.data.serialize as data_serialize
import imitation.policies.serialize as policies_serialize
from imitation.data import rollout, wrappers
from imitation.rewards.reward_wrapper import RewardVecEnvWrapper
from imitation.rewards.serialize import load_reward
from imitation.scripts.config.train_rl import train_rl_ex
from imitation.scripts.ingredients import environment
from imitation.scripts.ingredients import logging as logging_ingredient
from imitation.scripts.ingredients import policy_evaluation, rl


@train_rl_ex.main
def train_rl(
    *,
    total_timesteps: int,
    normalize_reward: bool,
    normalize_kwargs: dict,
    reward_type: Optional[str],
    reward_path: Optional[str],
    load_reward_kwargs: Optional[Mapping[str, Any]],
    rollout_save_final: bool,
    rollout_save_n_timesteps: Optional[int],
    rollout_save_n_episodes: Optional[int],
    policy_save_interval: int,
    policy_save_final: bool,
    agent_path: Optional[str],
    _rnd: np.random.Generator,
) -> Mapping[str, float]:
    """Trains an expert policy from scratch and saves the rollouts and policy.

    Checkpoints:
      At applicable training steps `step` (where step is either an integer or
      "final"):

        - Policies are saved to `{log_dir}/policies/{step}/`.
        - Rollouts are saved to `{log_dir}/rollouts/{step}.npz`.

    Args:
        total_timesteps: Number of training timesteps in `model.learn()`.
        normalize_reward: Applies normalization and clipping to the reward function by
            keeping a running average of training rewards. Note: this is may be
            redundant if using a learned reward that is already normalized.
        normalize_kwargs: kwargs for `VecNormalize`.
        reward_type: If provided, then load the serialized reward of this type,
            wrapping the environment in this reward. This is useful to test
            whether a reward model transfers. For more information, see
            `imitation.rewards.serialize.load_reward`.
        reward_path: A specifier, such as a path to a file on disk, used by
            reward_type to load the reward model. For more information, see
            `imitation.rewards.serialize.load_reward`.
        load_reward_kwargs: Additional kwargs to pass to `predict_processed`.
            Examples are 'alpha' for :class: `AddSTDRewardWrapper` and 'update_stats'
            for :class: `NormalizedRewardNet`.
        rollout_save_final: If True, then save rollouts right after training is
            finished.
        rollout_save_n_timesteps: The minimum number of timesteps saved in every
            file. Could be more than `rollout_save_n_timesteps` because
            trajectories are saved by episode rather than by transition.
            Must set exactly one of `rollout_save_n_timesteps`
            and `rollout_save_n_episodes`.
        rollout_save_n_episodes: The number of episodes saved in every
            file. Must set exactly one of `rollout_save_n_timesteps` and
            `rollout_save_n_episodes`.
        policy_save_interval: The number of training updates between in between
            intermediate rollout saves. If the argument is nonpositive, then
            don't save intermediate updates.
        policy_save_final: If True, then save the policy right after training is
            finished.
        agent_path: Path to load warm-started agent.
        _rnd: Random number generator provided by Sacred.

    Returns:
        The return value of `rollout_stats()` using the final policy.
    """
    custom_logger, log_dir = logging_ingredient.setup_logging()
    rollout_dir = log_dir / "rollouts"
    policy_dir = log_dir / "policies"
    rollout_dir.mkdir(parents=True, exist_ok=True)
    policy_dir.mkdir(parents=True, exist_ok=True)

    post_wrappers = [lambda env, idx: wrappers.RolloutInfoWrapper(env)]
    with environment.make_venv(  # type: ignore[wrong-keyword-args]
        post_wrappers=post_wrappers,
    ) as venv:
        callback_objs = []
        if reward_type is not None:
            reward_fn = load_reward(
                reward_type,
                reward_path,
                venv,
                **load_reward_kwargs,
            )
            venv = RewardVecEnvWrapper(venv, reward_fn)
            callback_objs.append(venv.make_log_callback())
            logging.info(f"Wrapped env in reward {reward_type} from {reward_path}.")

        if normalize_reward:
            # Normalize reward. Reward scale effectively changes the learning rate,
            # so normalizing it makes training more stable. Note we do *not* normalize
            # observations here; use the `NormalizeFeaturesExtractor` instead.
            venv = VecNormalize(venv, norm_obs=False, **normalize_kwargs)
            if reward_type == "RewardNet_normalized":
                warnings.warn(
                    "Applying normalization to already normalized reward function. \
                    Consider setting normalize_reward as False",
                    RuntimeWarning,
                )

        if policy_save_interval > 0:
            save_policy_callback: callbacks.EventCallback = (
                policies_serialize.SavePolicyCallback(policy_dir)
            )
            save_policy_callback = callbacks.EveryNTimesteps(
                policy_save_interval,
                save_policy_callback,
            )
            callback_objs.append(save_policy_callback)
        callback = callbacks.CallbackList(callback_objs)

        if agent_path is None:
            rl_algo = rl.make_rl_algo(venv)
        else:
            rl_algo = rl.load_rl_algo_from_path(agent_path=agent_path, venv=venv)
        rl_algo.set_logger(custom_logger)
        rl_algo.learn(total_timesteps, callback=callback)

        # Save final artifacts after training is complete.
        if rollout_save_final:
            save_path = rollout_dir / "final.npz"
            sample_until = rollout.make_sample_until(
                rollout_save_n_timesteps,
                rollout_save_n_episodes,
            )
            data_serialize.save(
                save_path,
                rollout.rollout(rl_algo, rl_algo.get_env(), sample_until, rng=_rnd),
            )
        if policy_save_final:
            output_dir = policy_dir / "final"
            policies_serialize.save_stable_model(output_dir, rl_algo)

        # Final evaluation of expert policy.
        eval_stats = policy_evaluation.eval_policy(rl_algo, venv)
        return eval_stats


def main_console():
    observer_path = pathlib.Path.cwd() / "output" / "sacred" / "train_rl"
    observer = FileStorageObserver(observer_path)
    train_rl_ex.observers.append(observer)
    train_rl_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



"""Utilities to parse sacred run directories."""
import json
import pathlib
import warnings
from collections import defaultdict
from typing import Any, Dict, Generator, List, Tuple

SacredRun = Dict[str, Any]
SacredConfAndRun = Tuple[Dict[str, Any], SacredRun]
GroupedRuns = Dict[str, Dict[str, List[SacredRun]]]


def find_sacred_runs(
    run_path: pathlib.Path,
    only_completed_runs: bool = False,
) -> Generator[SacredConfAndRun, None, None]:
    """Recursively iterates the sacred runs found below the given path.

    Assumes runs in the format of the sacred FileStorageObserver: each run consists
    of a folder that contains a config.json and a run.json file.

    Note: will work with nested directories and can therefore be applied to the
    `output/sacred` folder of the command line interface which creates sub-folders for
    each script.

    Args:
        run_path: The path to search for sacred run directories.
        only_completed_runs: If True, only yields runs that have a run.json file with
            status "COMPLETED".

    Yields:
        Tuples of (config, run) dicts.
    """
    for config_path in run_path.rglob("config.json"):
        run_path = config_path.parent / "run.json"

        if run_path.exists():
            run = json.loads(run_path.read_text())
            if only_completed_runs and run["status"] != "COMPLETED":
                continue
            conf = json.loads(config_path.read_text())
            yield conf, run
        else:  # pragma: no cover
            warnings.warn(f"Run {config_path.parent} has no run.json")


def group_runs_by_algo_and_env(
    path: pathlib.Path,
    only_completed_runs: bool = False,
) -> GroupedRuns:
    """Groups the runs found below the given path by algorithm and environment.

    Access all the runs of algorithm `algo` and environment `env` via
    `runs_by_algo_and_env[algo][env]`.

    Args:
        path: The path to search for sacred run directories.
        only_completed_runs: If True, only yields runs that have a run.json file with
            status "COMPLETED".

    Returns:
        A dictionary mapping algorithms to environments to lists of runs.
    """
    runs_by_algo_and_env: GroupedRuns = defaultdict(lambda: defaultdict(list))
    for conf, run in find_sacred_runs(path, only_completed_runs):
        algo = run["command"]
        env = conf["environment"]["gym_id"]
        runs_by_algo_and_env[algo][env].append(run)

    return runs_by_algo_and_env



"""Tests for `imitation.algorithms.density_baselines`."""

from dataclasses import asdict
from typing import Sequence, cast

import gymnasium as gym
import numpy as np
import pytest
import stable_baselines3
from stable_baselines3.common import policies, vec_env

from imitation.algorithms.density import DensityAlgorithm, DensityType
from imitation.data import rollout, types
from imitation.data.types import TrajectoryWithRew
from imitation.policies.base import RandomPolicy
from imitation.testing import reward_improvement

parametrize_density_stationary = pytest.mark.parametrize(
    "density_type,is_stationary",
    [(density_type, True) for density_type in DensityType]
    + [(DensityType.STATE_DENSITY, False)],
)


def score_trajectories(
    trajectories: Sequence[types.Trajectory],
    density_reward: DensityAlgorithm,
):
    # score trajectories under given reward function w/o discount
    returns = []
    for traj in trajectories:
        dones = np.zeros(len(traj), dtype=bool)
        dones[-1] = True
        steps = np.arange(0, len(traj.acts))
        obs = types.assert_not_dictobs(traj.obs)
        rewards = density_reward(obs[:-1], traj.acts, obs[1:], dones, steps)
        ret = np.sum(rewards)
        returns.append(ret)
    return returns


# test on Pendulum rather than Cartpole because I don't handle episodes that
# terminate early yet (see issue #40)
@parametrize_density_stationary
def test_density_reward(
    density_type,
    is_stationary,
    pendulum_venv,
    pendulum_expert_trajectories: Sequence[TrajectoryWithRew],
    rng,
):
    # use only a subset of trajectories
    n_experts = len(pendulum_expert_trajectories)
    expert_trajectories_train = pendulum_expert_trajectories[: n_experts // 2]
    reward_fn = DensityAlgorithm(
        demonstrations=expert_trajectories_train,
        density_type=density_type,
        kernel="gaussian",
        venv=pendulum_venv,
        is_stationary=is_stationary,
        kernel_bandwidth=0.2,
        standardise_inputs=True,
        rng=rng,
    )
    reward_fn.train()

    # check that expert policy does better than a random policy under our reward
    # function
    random_policy = RandomPolicy(
        pendulum_venv.observation_space,
        pendulum_venv.action_space,
    )
    sample_until = rollout.make_min_episodes(15)
    random_trajectories = rollout.generate_trajectories(
        random_policy,
        pendulum_venv,
        sample_until=sample_until,
        rng=rng,
    )
    expert_trajectories_test = pendulum_expert_trajectories[n_experts // 2 :]
    random_returns = score_trajectories(random_trajectories, reward_fn)
    expert_returns = score_trajectories(expert_trajectories_test, reward_fn)
    assert reward_improvement.is_significant_reward_improvement(
        random_returns,
        expert_returns,
    )


@pytest.mark.expensive
def test_density_trainer_smoke(
    pendulum_venv,
    pendulum_expert_trajectories: Sequence[TrajectoryWithRew],
    rng,
):
    # tests whether density trainer runs, not whether it's good
    # (it's actually really poor)
    rollouts = pendulum_expert_trajectories[:2]
    rl_algo = stable_baselines3.PPO(policies.ActorCriticPolicy, pendulum_venv)
    density_trainer = DensityAlgorithm(
        demonstrations=rollouts,
        venv=pendulum_venv,
        rl_algo=rl_algo,
        rng=rng,
    )
    density_trainer.train()
    density_trainer.train_policy(n_timesteps=2)
    density_trainer.test_policy(n_trajectories=2)


def test_density_with_other_trajectory_types(
    pendulum_expert_trajectories: Sequence[TrajectoryWithRew],
    pendulum_venv,
    rng,
):
    rl_algo = stable_baselines3.PPO(
        policies.ActorCriticPolicy,
        pendulum_venv,
        n_steps=10,  # small value to make test faster
        n_epochs=2,  # small value to make test faster
    )
    rollouts = pendulum_expert_trajectories[:2]
    transitions = rollout.flatten_trajectories_with_rew(rollouts)
    transitions_mappings = [cast(types.TransitionMapping, asdict(transitions))]
    d = DensityAlgorithm(
        demonstrations=transitions_mappings,
        venv=pendulum_venv,
        rl_algo=rl_algo,
        rng=rng,
    )
    d.train()
    d.train_policy(n_timesteps=2)
    d.test_policy(n_trajectories=2)

    minimal_transitions = types.TransitionsMinimal(
        obs=transitions.obs,
        acts=transitions.acts,
        infos=transitions.infos,
    )
    d = DensityAlgorithm(
        demonstrations=minimal_transitions,
        venv=pendulum_venv,
        rl_algo=rl_algo,
        rng=rng,
    )
    d.train()
    d.train_policy(n_timesteps=2)
    d.test_policy(n_trajectories=2)


def test_density_trainer_raises(
    pendulum_venv,
    rng,
):
    rl_algo = stable_baselines3.PPO(policies.ActorCriticPolicy, pendulum_venv)
    density_trainer = DensityAlgorithm(
        venv=pendulum_venv,
        rl_algo=rl_algo,
        rng=rng,
        demonstrations=None,
        density_type=DensityType.STATE_STATE_DENSITY,
    )
    with pytest.raises(ValueError, match="STATE_STATE_DENSITY requires next_obs_b"):
        density_trainer._get_demo_from_batch(
            np.zeros((1, 3)),
            np.zeros((1, 1)),
            None,
        )

    with pytest.raises(TypeError, match="Unsupported demonstration type"):
        density_trainer.set_demonstrations("foo")  # type: ignore[arg-type]


def test_dict_space(multi_obs_venv: vec_env.VecEnv):
    # multi-input policy to accept dict observations
    assert isinstance(multi_obs_venv.observation_space, gym.spaces.Dict)
    rl_algo = stable_baselines3.PPO(
        policies.MultiInputActorCriticPolicy,
        multi_obs_venv,
        n_steps=10,  # small value to make test faster
        n_epochs=2,  # small value to make test faster
    )
    rng = np.random.default_rng()

    # sample random transitions
    sample_until = rollout.make_min_episodes(15)
    rollouts = rollout.rollout(
        policy=None,
        venv=multi_obs_venv,
        sample_until=sample_until,
        rng=rng,
    )
    density_trainer = DensityAlgorithm(
        demonstrations=rollouts,
        kernel="gaussian",
        venv=multi_obs_venv,
        rl_algo=rl_algo,
        kernel_bandwidth=0.2,
        standardise_inputs=True,
        rng=rng,
        # SimpleMultiObsEnv has early stop (issue #40)
        allow_variable_horizon=True,
    )
    # confirm that training works
    density_trainer.train()
    density_trainer.train_policy(n_timesteps=2)
    density_trainer.test_policy(n_trajectories=2)



"""Tests ExplorationWrapper."""

import numpy as np
import pytest
import seals  # noqa: F401

from imitation.policies import exploration_wrapper
from imitation.util import util


def constant_policy(obs, state, mask):
    del state, mask  # Unused
    return np.zeros(len(obs), dtype=int), None


def fake_stateful_policy(obs, state, mask):
    del state, mask  # Unused
    return np.zeros(len(obs), dtype=int), (np.zeros(1),)


def make_wrapper(random_prob, switch_prob, rng):
    venv = util.make_vec_env(
        "seals/CartPole-v0",
        n_envs=1,
        rng=rng,
    )
    return (
        exploration_wrapper.ExplorationWrapper(
            policy=constant_policy,
            venv=venv,
            random_prob=random_prob,
            switch_prob=switch_prob,
            rng=rng,
        ),
        venv,
    )


def test_random_prob(rng):
    """Test that `random_prob` produces right behaviors of policy switching.

    The policy always makes an initial switch when ExplorationWrapper is applied.
    Then the policy makes switches according to `switch_prob`.

    This test fixes `switch_prob` to 0.5 and sets `random_prob` to 0.0, 1.0 and 0.5.
    (1) `random_prob=0.0`: Initial and following policies are always constant policies.
    (2) `random_prob=1.0`: Initial and following policies are always random policies.
    (3) `random_prob=0.5`: Around half-half for constant and random policies.

    Args:
        rng (np.random.Generator): random number generator.

    Raises:
        ValueError: Unknown policy type to switch.
    """
    wrapper, _ = make_wrapper(random_prob=0.0, switch_prob=0.5, rng=rng)
    assert wrapper.current_policy == constant_policy
    for _ in range(100):
        wrapper._switch()
        assert wrapper.current_policy == constant_policy

    wrapper, _ = make_wrapper(random_prob=1.0, switch_prob=0.5, rng=rng)
    assert wrapper.current_policy == wrapper._random_policy
    for _ in range(100):
        wrapper._switch()
        assert wrapper.current_policy == wrapper._random_policy

    wrapper, _ = make_wrapper(random_prob=0.5, switch_prob=0.5, rng=rng)
    num_random = 0
    num_constant = 0
    for _ in range(1000):
        wrapper._switch()
        if wrapper.current_policy == wrapper._random_policy:
            num_random += 1
        elif wrapper.current_policy == constant_policy:
            num_constant += 1
        else:  # pragma: no cover
            raise ValueError("Unknown policy")
    # Holds with very high probability (and seeding means it's deterministic)
    assert num_random > 450
    assert num_constant > 450


def test_switch_prob(rng):
    """Test that `switch_prob` produces right behaviors of policy switching.

    The policy always makes an initial switch when ExplorationWrapper is applied.
    Then the policy makes switches according to `switch_prob`.

    This test set includes the following:
    (1) `switch_prob=0.0`: The policy never switches after initial switch.
    (2) `switch_prob=1.0`: The policy always switches and the distribution of
        policies is determined by `random_prob`.

    Args:
        rng (np.random.Generator): random number generator.
    """
    wrapper, venv = make_wrapper(random_prob=0.5, switch_prob=0.0, rng=rng)
    policy = wrapper.current_policy

    obs = np.random.rand(100, 2)
    for action in wrapper(obs, None, None)[0]:
        assert venv.action_space.contains(action)
        assert wrapper.current_policy == policy

    def _always_switch(random_prob, num_steps):
        wrapper, _ = make_wrapper(random_prob=random_prob, switch_prob=1.0, rng=rng)
        num_random = 0
        num_constant = 0
        for _ in range(num_steps):
            obs = np.random.rand(1, 2)
            wrapper(obs, None, None)
            if wrapper.current_policy == wrapper._random_policy:
                num_random += 1
            elif wrapper.current_policy == constant_policy:
                num_constant += 1
            else:  # pragma: no cover
                raise ValueError("Unknown policy")
        return num_random, num_constant

    num_random, num_constant = _always_switch(
        random_prob=1.0,
        num_steps=5000,
    )
    assert num_random == 5000
    assert num_constant == 0
    num_random, num_constant = _always_switch(
        random_prob=0.5,
        num_steps=5000,
    )
    assert num_random > 2250
    assert num_constant > 2250
    num_random, num_constant = _always_switch(
        random_prob=0.0,
        num_steps=5000,
    )
    assert num_random == 0
    assert num_constant == 5000


def test_valid_output(rng):
    """Ensure that we test both the random and the wrapped policy at least once."""
    for random_prob in [0.0, 0.5, 1.0]:
        wrapper, venv = make_wrapper(random_prob=random_prob, switch_prob=0.5, rng=rng)
        np.random.seed(0)
        obs = np.random.rand(100, 2)
        for action in wrapper(obs, None, None)[0]:
            assert venv.action_space.contains(action)


def test_throws_for_stateful_policy(rng):
    venv = util.make_vec_env(
        "seals/CartPole-v0",
        n_envs=1,
        rng=rng,
    )
    wrapper = exploration_wrapper.ExplorationWrapper(
        policy=fake_stateful_policy,
        venv=venv,
        random_prob=0,
        switch_prob=0,
        rng=rng,
    )

    np.random.seed(0)
    obs = np.random.rand(100, 2)
    with pytest.raises(
        ValueError,
        match="Exploration wrapper does not support stateful policies.",
    ):
        wrapper(obs, (np.ones_like(obs)), None)

    with pytest.raises(
        ValueError,
        match="Exploration wrapper does not support stateful policies.",
    ):
        wrapper(obs, None, None)



"""Utility functions used to check if rewards improved wrt to previous rewards."""
from typing import Iterable

import numpy as np
from scipy import stats


def is_significant_reward_improvement(
    old_rewards: Iterable[float],
    new_rewards: Iterable[float],
    p_value: float = 0.05,
) -> bool:
    """Checks if the new rewards are really better than the old rewards.

    Ensures that this is not just due to lucky sampling by a permutation test.

    Args:
        old_rewards: Iterable of "old" trajectory rewards (e.g. before training).
        new_rewards: Iterable of "new" trajectory rewards (e.g. after training).
        p_value: The maximum probability, that the old rewards are just as good as the
            new rewards, that we tolerate.

    Returns:
        True, if the new rewards are most probably better than the old rewards.
        For this, the probability, that the old rewards are just as good as the new
        rewards must be below `p_value`.

    >>> is_significant_reward_improvement((5, 6, 7, 4, 4), (7, 5, 9, 9, 12))
    True

    >>> is_significant_reward_improvement((5, 6, 7, 4, 4), (7, 5, 9, 7, 4))
    False

    >>> is_significant_reward_improvement((5, 6, 7, 4, 4), (7, 5, 9, 7, 4), p_value=0.3)
    True
    """
    permutation_test_result = stats.permutation_test(
        (old_rewards, new_rewards),
        statistic=lambda x, y, axis: np.mean(x, axis=axis) - np.mean(y, axis=axis),
        vectorized=True,
        alternative="less",
    )

    return permutation_test_result.pvalue < p_value


def mean_reward_improved_by(
    old_rews: Iterable[float],
    new_rews: Iterable[float],
    min_improvement: float,
):
    """Checks if mean rewards improved wrt. to old rewards by a certain amount.

    Args:
        old_rews: Iterable of "old" trajectory rewards (e.g. before training).
        new_rews: Iterable of "new" trajectory rewards (e.g. after training).
        min_improvement: The minimum amount of improvement that we expect.

    Returns:
        True if the mean of the new rewards is larger than the mean of the old rewards
        by min_improvement.

    >>> mean_reward_improved_by([5, 8, 7], [8, 9, 10], 2)
    True

    >>> mean_reward_improved_by([5, 8, 7], [8, 9, 10], 5)
    False
    """
    improvement = np.mean(new_rews) - np.mean(old_rews)  # type: ignore[call-overload]
    return improvement >= min_improvement



"""Tests for `imitation.regularization.*`."""
import itertools
import tempfile

import numpy as np
import pytest
import torch as th

from imitation.regularization import regularizers, updaters
from imitation.util import logger as imit_logger


@pytest.fixture(
    scope="module",
    params=[
        (0.5, (0.9, 1)),  # unlikely to fall inside the interval
        (0.5, (0.01, 10)),  # likely to fall inside the interval
    ],
)
def interval_param_scaler(request):
    return updaters.IntervalParamScaler(*request.param)


@pytest.mark.parametrize(
    "lambda_",
    [
        10.0,
        0.001,
    ],
)
@pytest.mark.parametrize(
    "train_loss",
    [
        th.tensor(100.0),
        th.tensor(10.0),
        th.tensor(0.1),
        th.tensor(0.0),
        100.0,
        10.0,
        0.1,
        0.0,
    ],
)
def test_interval_param_scaler(lambda_, train_loss, interval_param_scaler):
    scaler = interval_param_scaler
    tolerable_interval = scaler.tolerable_interval
    scaling_factor = scaler.scaling_factor
    eps = np.finfo(float).eps
    if train_loss > eps:
        # The loss is a non-zero scalar, so we can construct a validation loss for
        # three different cases:

        # case that the ratio between the validation loss and the training loss is
        # above the tolerable interval
        val_loss = train_loss * tolerable_interval[1] * 2
        assert scaler(lambda_, train_loss, val_loss) == lambda_ * (1 + scaling_factor)

        # case that the ratio between the validation loss and the training loss is
        # below the tolerable interval
        val_loss = train_loss * tolerable_interval[0] / 2
        assert scaler(lambda_, train_loss, val_loss) == lambda_ * (1 - scaling_factor)

        # case that the ratio between the validation loss and the training loss is
        # within the tolerable interval
        val_loss = train_loss * (tolerable_interval[0] + tolerable_interval[1]) / 2
        assert scaler(lambda_, train_loss, val_loss) == lambda_
    else:
        # we have a zero loss. We try two cases. When the validation loss is zero,
        # the ratio is undefined, so we should return the current lambda. When the
        # validation loss is nonzero, the ratio is infinite, so we should see the lambda
        # increase by the scaling factor.
        # We try it for both a tensor and a float value.
        val_loss = th.tensor(0.0)
        assert scaler(lambda_, train_loss, val_loss) == lambda_
        val_loss = 0.0
        assert scaler(lambda_, train_loss, val_loss) == lambda_
        val_loss = th.tensor(1.0)
        assert scaler(lambda_, train_loss, val_loss) == lambda_ * (1 + scaling_factor)
        val_loss = 1.0
        assert scaler(lambda_, train_loss, val_loss) == lambda_ * (1 + scaling_factor)


def test_interval_param_scaler_raises(interval_param_scaler):
    scaler = interval_param_scaler
    with pytest.raises(ValueError, match="val_loss must be a scalar"):
        scaler(1.0, 1.0, th.Tensor([3.0, 4.0]))
    with pytest.raises(ValueError, match="train_loss must be a scalar"):
        scaler(1.0, th.Tensor([1.0, 2.0]), 1.0)
    with pytest.raises(ValueError, match="train_loss must be a scalar"):
        scaler(1.0, "random value", th.tensor(1.0))
    with pytest.raises(ValueError, match="val_loss must be a scalar"):
        scaler(1.0, 1.0, "random value")
    with pytest.raises(ValueError, match="lambda_ must be a float"):
        scaler(th.tensor(1.0), 1.0, 1.0)
    with pytest.raises(ValueError, match="lambda_ must not be zero.*"):
        scaler(0.0, 1.0, 1.0)
    with pytest.raises(ValueError, match="lambda_ must be non-negative.*"):
        scaler(-1.0, 1.0, 1.0)
    with pytest.raises(ValueError, match="losses must be non-negative.*"):
        scaler(1.0, -1.0, 1.0)
    with pytest.raises(ValueError, match="losses must be non-negative.*"):
        scaler(1.0, 1.0, -1.0)


def test_interval_param_scaler_init_raises():
    # this validates the value of scaling_factor.
    interval_err_msg = r"scaling_factor must be in \(0, 1\) within machine precision."

    with pytest.raises(ValueError, match=interval_err_msg):
        # cannot be negative as this is counter-intuitive to
        # the direction of scaling (just use the reciprocal).
        updaters.IntervalParamScaler(-1, (0.1, 0.9))

    with pytest.raises(ValueError, match=interval_err_msg):
        # cannot be larger than one as this would make lambda
        # negative when scaling down.
        updaters.IntervalParamScaler(1.1, (0.1, 0.9))

    with pytest.raises(ValueError, match=interval_err_msg):
        # cannot be exactly zero, as this never changes the value
        # of lambda when scaling up.
        updaters.IntervalParamScaler(0.0, (0.1, 0.9))

    with pytest.raises(ValueError, match=interval_err_msg):
        # cannot be exactly one, as when lambda is scaled down
        # this brings it to zero.
        updaters.IntervalParamScaler(1.0, (0.1, 0.9))

    # an interval obviously needs two elements only.
    with pytest.raises(
        ValueError,
        match="tolerable_interval must be a tuple of length 2",
    ):
        updaters.IntervalParamScaler(0.5, (0.1, 0.9, 0.5))  # type: ignore[arg-type]
    with pytest.raises(
        ValueError,
        match="tolerable_interval must be a tuple of length 2",
    ):
        updaters.IntervalParamScaler(0.5, (0.1,))  # type: ignore[arg-type]

    # the first element of the interval must be at least 0.
    with pytest.raises(
        ValueError,
        match="tolerable_interval must be a tuple whose first element "
        "is at least 0.*",
    ):
        updaters.IntervalParamScaler(0.5, (-0.1, 0.9))

    # the second element of the interval must be greater than the first.
    with pytest.raises(
        ValueError,
        match="tolerable_interval must be a tuple.*the second "
        "element is greater than the first",
    ):
        updaters.IntervalParamScaler(0.5, (0.1, 0.05))


@pytest.fixture(scope="module")
def hierarchical_logger():
    tmpdir = tempfile.mkdtemp()
    return imit_logger.configure(tmpdir, ["tensorboard", "stdout", "csv"])


@pytest.fixture(scope="module", params=[0.1, 1.0, 10.0])
def simple_optimizer(request):
    return th.optim.Adam([th.tensor(request.param, requires_grad=True)], lr=0.1)


@pytest.fixture(scope="module", params=[0.1, 1.0, 10.0])
def initial_lambda(request):
    return request.param


class SimpleRegularizer(regularizers.Regularizer[None]):
    """A simple regularizer that does nothing."""

    def regularize_and_backward(self, loss: th.Tensor) -> None:
        pass  # pragma: no cover


def test_regularizer_init_no_crash(
    initial_lambda,
    hierarchical_logger,
    simple_optimizer,
    interval_param_scaler,
):
    SimpleRegularizer(
        initial_lambda=initial_lambda,
        optimizer=simple_optimizer,
        logger=hierarchical_logger,
        lambda_updater=interval_param_scaler,
        val_split=0.2,
    )

    SimpleRegularizer(
        initial_lambda=initial_lambda,
        optimizer=simple_optimizer,
        logger=hierarchical_logger,
        lambda_updater=None,
        val_split=None,
    )

    SimpleRegularizer.create(
        initial_lambda=initial_lambda,
        lambda_updater=interval_param_scaler,
        val_split=0.2,
    )(
        optimizer=simple_optimizer,
        logger=hierarchical_logger,
    )


@pytest.mark.parametrize(
    "val_split",
    [
        0.0,
        1.0,
        -10,
        10,
        "random value",
        10**-100,
    ],
)
def test_regularizer_init_raises_on_val_split(
    initial_lambda,
    hierarchical_logger,
    simple_optimizer,
    interval_param_scaler,
    val_split,
):
    val_split_err_msg = "val_split.*must be a float.*between.*"
    with pytest.raises(ValueError, match=val_split_err_msg):
        return SimpleRegularizer(
            initial_lambda=initial_lambda,
            optimizer=simple_optimizer,
            logger=hierarchical_logger,
            lambda_updater=interval_param_scaler,
            val_split=val_split,
        )


def test_regularizer_init_raises(
    initial_lambda,
    hierarchical_logger,
    simple_optimizer,
    interval_param_scaler,
):
    with pytest.raises(
        ValueError,
        match=".*do not pass.*parameter updater.*regularization strength.*non-zero",
    ):
        SimpleRegularizer(
            initial_lambda=0.0,
            optimizer=simple_optimizer,
            logger=hierarchical_logger,
            lambda_updater=None,
            val_split=0.2,
        )
    with pytest.raises(
        ValueError,
        match=".*pass.*parameter updater.*must.*specify.*validation split.*",
    ):
        SimpleRegularizer(
            initial_lambda=initial_lambda,
            optimizer=simple_optimizer,
            logger=hierarchical_logger,
            lambda_updater=interval_param_scaler,
            val_split=None,
        )
    with pytest.raises(
        ValueError,
        match=".*pass.*validation split.*must.*pass.*parameter updater.*",
    ):
        SimpleRegularizer(
            initial_lambda=initial_lambda,
            optimizer=simple_optimizer,
            logger=hierarchical_logger,
            lambda_updater=None,
            val_split=0.2,
        )


@pytest.mark.parametrize(
    "train_loss",
    [
        th.tensor(10.0),
        th.tensor(1.0),
        th.tensor(0.1),
        th.tensor(0.01),
    ],
)
def test_regularizer_update_params(
    initial_lambda,
    hierarchical_logger,
    simple_optimizer,
    interval_param_scaler,
    train_loss,
):
    regularizer = SimpleRegularizer(
        initial_lambda=initial_lambda,
        logger=hierarchical_logger,
        lambda_updater=interval_param_scaler,
        optimizer=simple_optimizer,
        val_split=0.1,
    )
    val_to_train_loss_ratio = interval_param_scaler.tolerable_interval[1] * 2
    val_loss = train_loss * val_to_train_loss_ratio
    assert regularizer.lambda_ == initial_lambda
    assert (
        hierarchical_logger.default_logger.name_to_value["regularization_lambda"]
        == initial_lambda
    )
    regularizer.update_params(train_loss, val_loss)
    expected_lambda_value = interval_param_scaler(initial_lambda, train_loss, val_loss)
    assert regularizer.lambda_ == expected_lambda_value
    assert expected_lambda_value != initial_lambda
    assert (
        hierarchical_logger.default_logger.name_to_value["regularization_lambda"]
        == expected_lambda_value
    )


class SimpleLossRegularizer(regularizers.LossRegularizer):
    """A simple loss regularizer.

    It multiplies the total loss by lambda_+1.
    """

    def _loss_penalty(self, loss: regularizers.Scalar) -> regularizers.Scalar:
        return loss * self.lambda_  # this multiplies the total loss by lambda_+1.


@pytest.mark.parametrize(
    "train_loss_base",
    [
        th.tensor(10.0),
        th.tensor(1.0),
        th.tensor(0.1),
        th.tensor(0.01),
    ],
)
def test_loss_regularizer(
    hierarchical_logger,
    simple_optimizer,
    initial_lambda,
    train_loss_base,
):
    regularizer = SimpleLossRegularizer(
        initial_lambda=initial_lambda,
        logger=hierarchical_logger,
        lambda_updater=None,
        optimizer=simple_optimizer,
    )
    loss_param = simple_optimizer.param_groups[0]["params"][0]
    train_loss = train_loss_base * loss_param
    regularizer.optimizer.zero_grad()
    regularized_loss = regularizer.regularize_and_backward(train_loss)
    assert th.allclose(regularized_loss.data, train_loss * (initial_lambda + 1))
    assert (
        hierarchical_logger.default_logger.name_to_value["regularized_loss"]
        == regularized_loss
    )
    assert th.allclose(loss_param.grad, train_loss_base * (initial_lambda + 1))


class SimpleWeightRegularizer(regularizers.WeightRegularizer):
    """A simple weight regularizer.

    It multiplies the total weight by lambda_+1.
    """

    def _weight_penalty(self, weight, group):
        # this multiplies the total weight by lambda_+1.
        # However, the grad is only calculated with respect to the
        # previous value of the weight.
        # This difference is only noticeable if the grad of the loss
        # has a functional dependence on the weight (i.e. not linear).
        return weight * self.lambda_


@pytest.mark.parametrize(
    "train_loss_base",
    [
        th.tensor(10.0),
        th.tensor(1.0),
        th.tensor(0.1),
        th.tensor(0.01),
    ],
)
def test_weight_regularizer(
    hierarchical_logger,
    simple_optimizer,
    initial_lambda,
    train_loss_base,
):
    regularizer = SimpleWeightRegularizer(
        initial_lambda=initial_lambda,
        logger=hierarchical_logger,
        lambda_updater=None,
        optimizer=simple_optimizer,
    )
    weight = simple_optimizer.param_groups[0]["params"][0]
    initial_weight_value = weight.data.clone()
    regularizer.optimizer.zero_grad()
    train_loss = train_loss_base * th.pow(weight, 2) / 2
    regularizer.regularize_and_backward(train_loss)
    assert th.allclose(weight.data, initial_weight_value * (initial_lambda + 1))
    assert th.allclose(weight.grad, train_loss_base * initial_weight_value)


@pytest.mark.parametrize("p", [0.5, 1.5, -1, 0, "random value"])
def test_lp_regularizer_p_value_raises(hierarchical_logger, simple_optimizer, p):
    with pytest.raises(ValueError, match="p must be a positive integer"):
        regularizers.LpRegularizer(
            initial_lambda=1.0,
            logger=hierarchical_logger,
            lambda_updater=None,
            optimizer=simple_optimizer,
            p=p,
        )


MULTI_PARAM_OPTIMIZER_INIT_VALS = [-1.0, 0.0, 1.0]
MULTI_PARAM_OPTIMIZER_ARGS = itertools.product(
    MULTI_PARAM_OPTIMIZER_INIT_VALS,
    MULTI_PARAM_OPTIMIZER_INIT_VALS,
)


@pytest.fixture(scope="module", params=MULTI_PARAM_OPTIMIZER_ARGS)
def multi_param_optimizer(request):
    return th.optim.Adam(
        [th.tensor(p, requires_grad=True) for p in request.param],
        lr=0.1,
    )


MULTI_PARAM_AND_LR_OPTIMIZER_ARGS = itertools.product(
    MULTI_PARAM_OPTIMIZER_INIT_VALS,
    MULTI_PARAM_OPTIMIZER_INIT_VALS,
    [0.001, 0.01, 0.1],
)


@pytest.fixture(scope="module", params=MULTI_PARAM_AND_LR_OPTIMIZER_ARGS)
def multi_param_and_lr_optimizer(request):
    return th.optim.Adam(
        [th.tensor(p, requires_grad=True) for p in request.param[:-1]],
        lr=request.param[-1],
    )


@pytest.mark.parametrize(
    "train_loss",
    [
        th.tensor(10.0),
        th.tensor(1.0),
        th.tensor(0.1),
    ],
)
@pytest.mark.parametrize("p", [1, 2, 3])
def test_lp_regularizer(
    hierarchical_logger,
    multi_param_optimizer,
    initial_lambda,
    train_loss,
    p,
):
    regularizer = regularizers.LpRegularizer(
        initial_lambda=initial_lambda,
        logger=hierarchical_logger,
        lambda_updater=None,
        optimizer=multi_param_optimizer,
        p=p,
    )
    params = multi_param_optimizer.param_groups[0]["params"]
    regularizer.optimizer.zero_grad()
    regularized_loss = regularizer.regularize_and_backward(train_loss)
    loss_penalty = sum(
        [th.linalg.vector_norm(param.data, ord=p).pow(p) for param in params],
    )
    assert th.allclose(
        regularized_loss.data,
        train_loss + initial_lambda * loss_penalty,
    )
    assert (
        regularized_loss
        == hierarchical_logger.default_logger.name_to_value["regularized_loss"]
    )
    for param in params:
        assert th.allclose(
            param.grad,
            p * initial_lambda * th.abs(param).pow(p - 1) * th.sign(param),
        )


@pytest.mark.parametrize(
    "train_loss_base",
    [
        th.tensor(1.0),
        th.tensor(0.1),
        th.tensor(0.01),
    ],
)
def test_weight_decay_regularizer(
    multi_param_and_lr_optimizer,
    hierarchical_logger,
    initial_lambda,
    train_loss_base,
):
    regularizer = regularizers.WeightDecayRegularizer(
        initial_lambda=initial_lambda,
        logger=hierarchical_logger,
        lambda_updater=None,
        optimizer=multi_param_and_lr_optimizer,
    )
    weights = regularizer.optimizer.param_groups[0]["params"]
    lr = regularizer.optimizer.param_groups[0]["lr"]
    initial_weight_values = [weight.data.clone() for weight in weights]
    regularizer.optimizer.zero_grad()
    train_loss = train_loss_base * sum(th.pow(weight, 2) / 2 for weight in weights)
    regularizer.regularize_and_backward(train_loss)
    for weight, initial_weight_value in zip(weights, initial_weight_values):
        assert th.allclose(
            weight.data,
            initial_weight_value * (1 - lr * initial_lambda),
        )
        assert th.allclose(weight.grad, train_loss_base * initial_weight_value)



"""imitation: implementations of imitation and reward learning algorithms."""

from importlib import metadata

try:
    __version__ = metadata.version("imitation")
except metadata.PackageNotFoundError:  # pragma: no cover
    # package is not installed
    pass



"""Configuration settings for analyze, inspecting results from completed experiments."""

import os.path as osp

import sacred

analysis_ex = sacred.Experiment("analyze")


@analysis_ex.config
def config():
    # Recursively search in this directory for sacred logs
    source_dir_str = "output/sacred/train_adversarial"
    skip_failed_runs = True  # Skip analysis for logs that have FAILED status
    run_name = None  # Restrict analysis to sacred logs with a certain run name
    env_name = None  # Restrict analysis to sacred logs with a certain env name
    csv_output_path = None  # Write CSV output to this path
    tex_output_path = None  # Write LaTex output to this path
    print_table = True  # Set to True to print analysis to stdout
    split_str = ","  # str used to split source_dir_str into multiple source dirs
    table_verbosity = 1  # Choose from 0, 1, 2 or 3
    source_dirs = None


@analysis_ex.config
def convert_source_dirs(source_dir_str, split_str, source_dirs):
    if source_dirs is None:
        source_dirs = source_dir_str.split(split_str)

    source_dirs = [osp.expanduser(p) for p in source_dirs]



"""Tests for `imitation.data.wrappers`."""

from typing import List, Sequence, Type

import gymnasium as gym
import numpy as np
import pytest
from stable_baselines3.common.vec_env import DummyVecEnv

from imitation.data import types
from imitation.data.wrappers import BufferingWrapper


class _CountingEnv(gym.Env):  # pragma: no cover
    """At timestep `t` of each episode, has `reward / 10 == obs == t`.

    Episodes finish after `episode_length` calls to `step()`. As an example,
    if we have `episode_length=5`, then an episode is to have the
    following observations and rewards:

    ```
    obs = [0, 1, 2, 3, 4, 5]
    rews = [10, 20, 30, 40, 50]
    ```
    """

    def __init__(self, episode_length=5):
        assert episode_length >= 1
        self.observation_space = gym.spaces.Box(low=0, high=np.inf, shape=())
        self.action_space = gym.spaces.Box(low=0, high=np.inf, shape=())
        self.episode_length = episode_length
        self.timestep = None

    def reset(self, seed=None):
        t, self.timestep = 0, 1
        return t, {}

    def step(self, action):
        if self.timestep is None:
            raise RuntimeError("Need to reset before first step().")
        if self.timestep > self.episode_length:
            raise RuntimeError("Episode is over. Need to step().")
        if np.array(action) not in self.action_space:
            raise ValueError(f"Invalid action {action}")

        t, self.timestep = self.timestep, self.timestep + 1
        done = t == self.episode_length
        return t, t * 10, done, False, {}


class _CountingDictEnv(_CountingEnv):  # pragma: no cover
    """Similar to _CountingEnv, but with Dict observation."""

    def __init__(self, episode_length=5):
        super().__init__(episode_length)
        self.observation_space = gym.spaces.Dict(
            spaces={"t": gym.spaces.Box(low=0, high=np.inf, shape=())},
        )

    def reset(self, seed=None):
        t, self.timestep = 0.0, 1.0
        return {"t": t}, {}

    def step(self, action):
        if self.timestep is None:
            raise RuntimeError("Need to reset before first step().")
        if self.timestep > self.episode_length:
            raise RuntimeError("Episode is over. Need to step().")
        if np.array(action) not in self.action_space:
            raise ValueError(f"Invalid action {action}")

        t, self.timestep = self.timestep, self.timestep + 1
        done = t == self.episode_length
        return {"t": t}, t * 10, done, False, {}


Envs = [_CountingEnv, _CountingDictEnv]


def _make_buffering_venv(
    Env: Type[gym.Env],
    error_on_premature_reset: bool,
) -> BufferingWrapper:
    venv = DummyVecEnv([Env] * 2)
    wrapped_venv = BufferingWrapper(venv, error_on_premature_reset)
    wrapped_venv.reset()
    return wrapped_venv


def _assert_equal_scrambled_vectors(a: np.ndarray, b: np.ndarray) -> None:
    """Raises AssertionError if `a` and `b` are not identical up to sorting."""
    assert a.shape == b.shape
    assert a.ndim == 1
    np.testing.assert_allclose(np.sort(a), np.sort(b))


def _join_transitions(
    trans_list: Sequence[types.TransitionsWithRew],
) -> types.TransitionsWithRew:
    def concat(x):
        return np.concatenate(list(x))

    obs = concat(t.obs for t in trans_list)
    next_obs = concat(t.next_obs for t in trans_list)
    rews = concat(t.rews for t in trans_list)
    acts = concat(t.acts for t in trans_list)
    dones = concat(t.dones for t in trans_list)
    infos = concat(t.infos for t in trans_list)
    return types.TransitionsWithRew(
        obs=obs,
        next_obs=next_obs,
        rews=rews,
        acts=acts,
        dones=dones,
        infos=infos,
    )


@pytest.mark.parametrize("Env", Envs)
@pytest.mark.parametrize("episode_lengths", [(1,), (6, 5, 1, 2), (2, 2)])
@pytest.mark.parametrize("n_steps", [1, 2, 20, 21])
@pytest.mark.parametrize("extra_pop_timesteps", [(), (1,), (4, 8)])
def test_pop(
    Env: Type[gym.Env],
    episode_lengths: Sequence[int],
    n_steps: int,
    extra_pop_timesteps: Sequence[int],
) -> None:
    """Check pop_transitions() results for BufferWrapper.

    To make things easier to test, we use _CountingEnv where the observation
    is simply the episode timestep. The reward is 10x the timestep. Our action
    is 2.1x the timestep. There is an confusing offset for the observation because
    it has timestep 0 (due to reset()) and the other quantities don't, so here is
    an example of environment outputs and associated actions:

    ```
    episode_length = 5
    obs = [0, 1, 2, 3, 4, 5]  (len=6)
    acts = [0, 2.1, 4.2, ..., 8.4]  (len=5)
    rews = [10, ..., 50]  (len=5)
    ```

    Converted to `Transition`-format, this looks like:
    ```
    episode_length = 5
    obs = [0, 1, 2, 3, 4, 5]  (len=5)
    next_obs = [1, 2, 3, 4, 5]  (len=5)
    acts = [0, 2.1, 4.2, ..., 8.4]  (len=5)
    rews = [10, ..., 50]  (len=5)
    ```

    Args:
        Env: Environment class type.
        episode_lengths: The number of timesteps before episode end in each dummy
            environment.
        n_steps: Number of times to call `step()` on the dummy environment.
        extra_pop_timesteps: By default, we only call `pop_*()` after `n_steps` calls
            to `step()`. For every unique positive `x` in `extra_pop_timesteps`, we
            also call `pop_*()` after the `x`th call to `step()`. All popped samples
            are concatenated before validating results at the end of this test case.
            All `x` in `extra_pop_timesteps` must be in range(1, n_steps). (`x == 0`
            is not valid because there are no transitions to pop at timestep 0).

    Raises:
        ValueError: `n_steps <= 0`.
    """
    if not n_steps >= 1:  # pragma: no cover
        raise ValueError(f"n_steps = {n_steps} <= 0")
    for t in extra_pop_timesteps:  # pragma: no cover
        if t < 1:
            raise ValueError(t)
        if not 1 <= t < n_steps:
            pytest.skip("pop timesteps out of bounds for this test case")

    def make_env(ep_len):
        return lambda: Env(episode_length=ep_len)

    venv = DummyVecEnv([make_env(ep_len) for ep_len in episode_lengths])
    venv_buffer = BufferingWrapper(venv)

    # To test `pop_transitions`, we will check that every obs, act, and rew
    # returned by `.reset()` and `.step()` is also returned by one of the
    # calls to `pop_transitions()`.
    transitions_list: List[types.TransitionsWithRew] = []

    # Initial observation (only matters for pop_transitions()).
    obs = venv_buffer.reset()
    if Env == _CountingEnv:
        np.testing.assert_array_equal(obs, [0] * venv.num_envs)
    else:
        np.testing.assert_array_equal(obs["t"], [0] * venv.num_envs)

    for t in range(1, n_steps + 1):
        acts = obs * 2.1 if Env == _CountingEnv else obs["t"] * 2.1
        venv_buffer.step_async(acts)
        obs, *_ = venv_buffer.step_wait()

        if t in extra_pop_timesteps:
            transitions_list.append(venv_buffer.pop_transitions())

    transitions_list.append(venv_buffer.pop_transitions())

    # Build expected transitions
    expect_obs_list = []
    for ep_len in episode_lengths:
        n_complete, remainder = divmod(n_steps, ep_len)
        expect_obs_list.extend([np.arange(ep_len)] * n_complete)
        expect_obs_list.append(np.arange(remainder))

    expect_obs = np.concatenate(expect_obs_list)
    expect_next_obs = expect_obs + 1
    expect_acts = expect_obs * 2.1
    expect_rews = expect_next_obs * 10

    # Check `pop_transitions()`
    trans = _join_transitions(transitions_list)
    if Env == _CountingEnv:
        actual_obs = types.assert_not_dictobs(trans.obs)
        actual_next_obs = types.assert_not_dictobs(trans.next_obs)
    else:
        actual_obs = types.DictObs.stack(trans.obs).get("t")
        actual_next_obs = types.DictObs.stack(trans.next_obs).get("t")
    _assert_equal_scrambled_vectors(actual_obs, expect_obs)
    _assert_equal_scrambled_vectors(actual_next_obs, expect_next_obs)
    _assert_equal_scrambled_vectors(trans.acts, expect_acts)
    _assert_equal_scrambled_vectors(trans.rews, expect_rews)


@pytest.mark.parametrize("Env", Envs)
def test_reset_error(Env: Type[gym.Env]):
    # Resetting before a `step()` is okay.
    for flag in [True, False]:
        venv = _make_buffering_venv(Env, flag)
        for _ in range(10):
            venv.reset()

    # Resetting after a `step()` is not okay if error flag is True.
    venv = _make_buffering_venv(Env, True)
    zeros = np.array([0.0, 0.0], dtype=venv.action_space.dtype)
    venv.step(zeros)
    with pytest.raises(RuntimeError, match="before samples were accessed"):
        venv.reset()

    # Same as previous case, but insert a `pop_transitions()` in between.
    venv = _make_buffering_venv(Env, True)
    venv.step(zeros)
    venv.pop_transitions()
    venv.step(zeros)
    with pytest.raises(RuntimeError, match="before samples were accessed"):
        venv.reset()

    # Resetting after a `step()` is ok if error flag is False.
    venv = _make_buffering_venv(Env, False)
    venv.step(zeros)
    venv.reset()

    # Resetting after a `step()` is ok if transitions are first collected.
    for flag in [True, False]:
        venv = _make_buffering_venv(Env, flag)
        venv.step(zeros)
        venv.pop_transitions()
        venv.reset()


@pytest.mark.parametrize("Env", Envs)
def test_n_transitions_and_empty_error(Env: Type[gym.Env]):
    venv = _make_buffering_venv(Env, True)
    trajs, ep_lens = venv.pop_trajectories()
    assert trajs == []
    assert ep_lens == []
    zeros = np.array([0.0, 0.0], dtype=venv.action_space.dtype)
    venv.step(zeros)
    assert venv.n_transitions == 2
    venv.step(zeros)
    assert venv.n_transitions == 4
    venv.pop_transitions()
    assert venv.n_transitions == 0
    with pytest.raises(RuntimeError, match=".* empty .*"):
        venv.pop_transitions()



"""This ingredient provides a SQIL algorithm instance."""
import sacred
from stable_baselines3 import dqn as dqn_algorithm

from imitation.policies import base
from imitation.scripts.ingredients import policy, rl

sqil_ingredient = sacred.Ingredient(
    "sqil",
    ingredients=[rl.rl_ingredient, policy.policy_ingredient],
)


@sqil_ingredient.config
def config():
    total_timesteps = 3e5
    train_kwargs = dict(
        log_interval=4,  # Number of updates between Tensorboard/stdout logs
        progress_bar=True,
    )

    locals()  # quieten flake8 unused variable warning


@rl.rl_ingredient.config_hook
def override_rl_cls(config, command_name, logger):
    # want to remove arguments added by the rl ingredient but keep
    # the ones that are added by others
    del logger

    res = {}
    if command_name == "sqil" and config["rl"]["rl_cls"] is None:
        res["rl_cls"] = dqn_algorithm.DQN

    return res


@policy.policy_ingredient.config_hook
def override_policy_cls(config, command_name, logger):  # noqa
    del logger

    res = {}
    if (
        command_name == "sqil"
        and config["policy"]["policy_cls"] == base.FeedForward32Policy
    ):
        res["policy_cls"] = "MlpPolicy"

    return res



#!/usr/bin/env python
"""Convert trajectories from `imitation` format to openai/baselines GAIL format."""

import argparse
import os
from pathlib import Path
from typing import Sequence

import numpy as np

from imitation.data import rollout, serialize, types


def convert_trajs_to_sb(trajs: Sequence[types.TrajectoryWithRew]) -> dict:
    """Converts Trajectories into the dict format used by Stable Baselines GAIL."""
    trans = rollout.flatten_trajectories_with_rew(trajs)
    return dict(
        acs=trans.acts,
        rews=trans.rews,
        obs=trans.obs,
        ep_rets=np.array([np.sum(t.rews) for t in trajs]),
    )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("src_path", type=str)
    parser.add_argument("dst_path", type=str)
    args = parser.parse_args()

    src_path = Path(args.src_path)
    dst_path = Path(args.dst_path)

    assert src_path.is_file()
    src_trajs = serialize.load_with_rewards(src_path)
    dst_trajs = convert_trajs_to_sb(src_trajs)
    os.makedirs(dst_path.parent, exist_ok=True)
    with open(dst_path, "wb") as f:
        np.savez_compressed(f, **dst_trajs)

    print(f"Dumped rollouts to {dst_path}")


if __name__ == "__main__":
    main()



"""Setup for imitation: a reward and imitation learning library."""

import os
import warnings
from sys import platform
from typing import TYPE_CHECKING

from setuptools import find_packages, setup
from setuptools.command.install import install

if TYPE_CHECKING:
    from setuptools_scm.version import ScmVersion

IS_NOT_WINDOWS = os.name != "nt"

PARALLEL_REQUIRE = ["ray[debug,tune]~=2.0.0"]
ATARI_REQUIRE = [
    "seals[atari]~=0.2.1",
]
PYTYPE = ["pytype==2023.9.27"] if IS_NOT_WINDOWS else []

HYPOTHESIS = ["hypothesis~=6.54.1"]

# Note: the versions of the test and doc requirements should be tightly pinned to known
#   working versions to make our CI/CD pipeline as stable as possible.
TESTS_REQUIRE = (
    [
        "black[jupyter]~=22.6.0",
        "coverage~=6.4.2",
        "codecov~=2.1.12",
        "codespell~=2.1.0",
        "darglint~=1.8.1",
        "filelock~=3.7.1",
        "flake8~=4.0.1",
        "flake8-blind-except==0.2.1",
        "flake8-builtins~=1.5.3",
        "flake8-commas~=2.1.0",
        "flake8-debugger~=4.1.2",
        "flake8-docstrings~=1.6.0",
        "flake8-isort~=4.1.2",
        "ipykernel~=6.15.1",
        "jupyter~=1.0.0",
        # TODO: upgrade jupyter-client once
        #  https://github.com/jupyter/jupyter_client/issues/637 is fixed
        "jupyter-client~=6.1.12",
        "moviepy~=1.0.3",
        "mypy~=0.990",
        "pandas~=1.4.3",
        "pytest~=7.1.2",
        "pytest-cov~=3.0.0",
        "pytest-notebook==0.8.0",
        "pytest-timeout~=2.1.0",
        "pytest-xdist~=2.5.0",
        "scipy~=1.9.0",
        "wandb==0.12.21",
        "setuptools_scm~=7.0.5",
        "pre-commit>=2.20.0",
    ]
    + PARALLEL_REQUIRE
    + ATARI_REQUIRE
    + PYTYPE
    + HYPOTHESIS
)
DOCS_REQUIRE = (
    [
        "sphinx~=5.1.1",
        "sphinx-autodoc-typehints~=1.19.1",
        "sphinx-rtd-theme~=1.0.0",
        "sphinxcontrib-napoleon==0.7",
        "furo==2022.6.21",
        "sphinx-copybutton==0.5.0",
        "sphinx-github-changelog~=1.2.0",
        "myst-nb==0.17.2",
        "ipykernel~=6.15.2",
    ]
    + ATARI_REQUIRE
    + PARALLEL_REQUIRE
    + HYPOTHESIS
)


def get_readme() -> str:
    """Retrieve content from README."""
    with open("README.md", "r", encoding="utf-8") as f:
        return f.read()


class InstallCommand(install):
    """Custom install command to throw warnings about external dependencies."""

    def run(self):
        """Run the install command."""
        install.run(self)

        if platform == "darwin":
            warnings.warn(
                "Installation of important packages for macOS is required. "
                "Scripts in the experiments folder will likely not run without these "
                "packages: gnu-getopt, parallel, coreutils. They can be installed with "
                "Homebrew by running `brew install gnu-getopt parallel coreutils`."
                "See https://brew.sh/ for installation instructions.",
            )


def get_version(version: "ScmVersion") -> str:
    """Generates the version string for the package.

    This function replaces the default version format used by setuptools_scm
    to allow development builds to be versioned using the git commit hash
    instead of the number of commits since the last release, which leads to
    duplicate version identifiers when using multiple branches
    (see https://github.com/HumanCompatibleAI/imitation/issues/500).

    The version has the following format:

    {version}[.dev{build}]
    where build is the shortened commit hash converted to base 10.

    Args:
        version: The version object given by setuptools_scm, calculated
            from the git repository.

    Returns:
        The formatted version string to use for the package.
    """
    # We import setuptools_scm here because it is only installed after the module
    # is loaded and the setup function is called.
    from setuptools_scm import version as scm_version

    if version.node:
        # By default node corresponds to the short commit hash when using git,
        # plus a "g" prefix. We remove the "g" prefix from the commit hash which
        # is added by setuptools_scm by default ("g" for git vs. mercurial etc.)
        # because letters are not valid for version identifiers in PEP 440.
        # We also convert from hexadecimal to base 10 for the same reason.
        version.node = str(int(version.node.lstrip("g"), 16))
    if version.exact:
        # an exact version is when the current commit is tagged with a version.
        return version.format_with("{tag}")
    else:
        # the current commit is not tagged with a version, so we guess
        # what the "next" version will be (this can be disabled but is the
        # default behavior of setuptools_scm so it has been left in).
        return version.format_next_version(
            scm_version.guess_next_version,
            fmt="{guessed}.dev{node}",
        )


def get_local_version(version: "ScmVersion", time_format="%Y%m%d") -> str:
    """Generates the local version string for the package.

    By default, when commits are made on top of a release version, setuptools_scm
    sets the version to be {version}.dev{distance}+{node} where {distance} is the number
    of commits since the last release and {node} is the short commit hash.
    This function replaces the default version format used by setuptools_scm
    so that committed changes away from a release version are not considered
    local versions but dev versions instead (by using the format
    {version}.dev{node} instead. This is so that we can push test releases
    to TestPyPI (it does not accept local versions).

    Local versions are still present if there are uncommitted changes (if the tree
    is dirty), in which case the current date is added to the version.

    Args:
        version: The version object given by setuptools_scm, calculated
            from the git repository.
        time_format: The format to use for the date.

    Returns:
        The formatted local version string to use for the package.
    """
    return version.format_choice(
        "",
        "+d{time:{time_format}}",
        time_format=time_format,
    )


setup(
    cmdclass={"install": InstallCommand},
    name="imitation",
    use_scm_version={"local_scheme": get_local_version, "version_scheme": get_version},
    setup_requires=["setuptools_scm"],
    description="Implementation of modern reward and imitation learning algorithms.",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    author="Center for Human-Compatible AI and Google",
    python_requires=">=3.8.0",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"imitation": ["py.typed", "scripts/config/tuned_hps/*.json"]},
    # Note: while we are strict with our test and doc requirement versions, we try to
    #   impose as little restrictions on the install requirements as possible. Try to
    #   encode only known incompatibilities here. This prevents nasty dependency issues
    #   for our users.
    install_requires=[
        "gymnasium[classic-control]~=0.29",
        "matplotlib",
        "numpy>=1.15",
        "torch>=1.4.0",
        "tqdm",
        "rich",
        "scikit-learn>=0.21.2",
        "seals~=0.2.1",
        "stable-baselines3~=2.2.1",
        "sacred>=0.8.4",
        "tensorboard>=1.14",
        "huggingface_sb3~=3.0",
        "optuna>=3.0.1",
        "datasets>=2.8.0",
    ],
    tests_require=TESTS_REQUIRE,
    extras_require={
        # recommended packages for development
        "dev": [
            "autopep8",
            "ipdb",
            "isort~=5.0",
            "codespell",
            "sphinx-autobuild",
            # for convenience
            *TESTS_REQUIRE,
            *DOCS_REQUIRE,
        ]
        + PYTYPE,
        "test": TESTS_REQUIRE,
        "docs": DOCS_REQUIRE,
        "parallel": PARALLEL_REQUIRE,
        "mujoco": [
            "gymnasium[classic-control,mujoco]~=0.29",
        ],
        "atari": ATARI_REQUIRE,
    },
    entry_points={
        "console_scripts": [
            "imitation-eval-policy=imitation.scripts.eval_policy:main_console",
            "imitation-parallel=imitation.scripts.parallel:main_console",
            (
                "imitation-train-adversarial="
                "imitation.scripts.train_adversarial:main_console"
            ),
            "imitation-train-imitation=imitation.scripts.train_imitation:main_console",
            (
                "imitation-train-preference-comparisons="
                "imitation.scripts.train_preference_comparisons:main_console"
            ),
            "imitation-train-rl=imitation.scripts.train_rl:main_console",
        ],
    },
    url="https://github.com/HumanCompatibleAI/imitation",
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



"""Trains DAgger on synthetic demonstrations generated from an expert policy."""

import logging
import os.path as osp
import pathlib
from typing import Any, Dict, Mapping, Optional, Sequence, cast

import numpy as np
from sacred.observers import FileStorageObserver

from imitation.algorithms import dagger as dagger_algorithm
from imitation.algorithms import sqil as sqil_algorithm
from imitation.data import rollout, types
from imitation.scripts.config.train_imitation import train_imitation_ex
from imitation.scripts.ingredients import bc as bc_ingredient
from imitation.scripts.ingredients import demonstrations, environment, expert
from imitation.scripts.ingredients import logging as logging_ingredient
from imitation.scripts.ingredients import policy_evaluation
from imitation.util import util

logger = logging.getLogger(__name__)


def _all_trajectories_have_reward(trajectories: Sequence[types.Trajectory]) -> bool:
    """Returns True if all trajectories have reward information."""
    return all(isinstance(t, types.TrajectoryWithRew) for t in trajectories)


def _try_computing_expert_stats(
    expert_trajs: Sequence[types.Trajectory],
) -> Optional[Mapping[str, float]]:
    """Adds expert statistics to `stats` if all expert trajectories have reward."""
    if _all_trajectories_have_reward(expert_trajs):
        return rollout.rollout_stats(
            cast(Sequence[types.TrajectoryWithRew], expert_trajs),
        )
    else:
        logger.warning(
            "Expert trajectories do not have reward information, so expert "
            "statistics cannot be computed.",
        )
        return None


def _collect_stats(
    imit_stats: Mapping[str, float],
    expert_trajs: Sequence[types.Trajectory],
) -> Mapping[str, Mapping[str, Any]]:
    stats = {"imit_stats": imit_stats}
    expert_stats = _try_computing_expert_stats(expert_trajs)
    if expert_stats is not None:
        stats["expert_stats"] = expert_stats

    return stats


@train_imitation_ex.command
def bc(
    bc: Dict[str, Any],
    _run,
    _rnd: np.random.Generator,
) -> Mapping[str, Mapping[str, float]]:
    """Runs BC training.

    Args:
        bc: Configuration for BC training.
        _run: Sacred run object.
        _rnd: Random number generator provided by Sacred.

    Returns:
        Statistics for rollouts from the trained policy and demonstration data.
    """
    custom_logger, log_dir = logging_ingredient.setup_logging()

    expert_trajs = demonstrations.get_expert_trajectories()
    with environment.make_venv() as venv:  # type: ignore[wrong-arg-count]
        bc_trainer = bc_ingredient.make_bc(venv, expert_trajs, custom_logger)

        bc_train_kwargs = dict(log_rollouts_venv=venv, **bc["train_kwargs"])
        if bc_train_kwargs["n_epochs"] is None and bc_train_kwargs["n_batches"] is None:
            bc_train_kwargs["n_batches"] = 50_000

        bc_trainer.train(**bc_train_kwargs)
        # TODO(adam): add checkpointing to BC?
        util.save_policy(bc_trainer.policy, policy_path=osp.join(log_dir, "final.th"))

        imit_stats = policy_evaluation.eval_policy(bc_trainer.policy, venv)

    stats = _collect_stats(imit_stats, expert_trajs)

    return stats


@train_imitation_ex.command
def dagger(
    bc: Dict[str, Any],
    dagger: Mapping[str, Any],
    _run,
    _rnd: np.random.Generator,
) -> Mapping[str, Mapping[str, float]]:
    """Runs DAgger training.

    Args:
        bc: Configuration for BC training.
        dagger: Arguments for DAgger training.
        _run: Sacred run object.
        _rnd: Random number generator provided by Sacred.

    Returns:
        Statistics for rollouts from the trained policy and demonstration data.
    """
    custom_logger, log_dir = logging_ingredient.setup_logging()

    expert_trajs: Optional[Sequence[types.Trajectory]] = None
    if dagger["use_offline_rollouts"]:
        expert_trajs = demonstrations.get_expert_trajectories()

    with environment.make_venv() as venv:  # type: ignore[wrong-arg-count]
        bc_trainer = bc_ingredient.make_bc(venv, expert_trajs, custom_logger)

        bc_train_kwargs = dict(log_rollouts_venv=venv, **bc["train_kwargs"])
        if bc_train_kwargs["n_epochs"] is None and bc_train_kwargs["n_batches"] is None:
            bc_train_kwargs["n_epochs"] = 4

        expert_policy = expert.get_expert_policy(venv)

        dagger_trainer = dagger_algorithm.SimpleDAggerTrainer(
            venv=venv,
            scratch_dir=osp.join(log_dir, "scratch"),
            expert_trajs=expert_trajs,
            expert_policy=expert_policy,
            custom_logger=custom_logger,
            bc_trainer=bc_trainer,
            beta_schedule=dagger["beta_schedule"],
            rng=_rnd,
        )

        dagger_trainer.train(
            total_timesteps=int(dagger["total_timesteps"]),
            bc_train_kwargs=bc_train_kwargs,
        )
        # TODO(adam): add checkpointing to DAgger?
        save_locations = dagger_trainer.save_trainer()
        print(f"Model saved to {save_locations}")

        imit_stats = policy_evaluation.eval_policy(bc_trainer.policy, venv)

    assert dagger_trainer._all_demos is not None
    stats = _collect_stats(imit_stats, dagger_trainer._all_demos)

    return stats


@train_imitation_ex.command
def sqil(
    sqil: Mapping[str, Any],
    policy: Mapping[str, Any],
    rl: Mapping[str, Any],
    _run,
    _rnd: np.random.Generator,
) -> Mapping[str, Mapping[str, float]]:
    custom_logger, log_dir = logging_ingredient.setup_logging()
    expert_trajs = demonstrations.get_expert_trajectories()

    with environment.make_venv() as venv:  # type: ignore[wrong-arg-count]
        sqil_trainer = sqil_algorithm.SQIL(
            venv=venv,
            demonstrations=expert_trajs,
            policy=policy["policy_cls"],
            custom_logger=custom_logger,
            rl_algo_class=rl["rl_cls"],
            rl_kwargs=rl["rl_kwargs"],
        )

        sqil_trainer.train(
            total_timesteps=int(sqil["total_timesteps"]),
            **sqil["train_kwargs"],
        )
        util.save_policy(sqil_trainer.policy, policy_path=osp.join(log_dir, "final.th"))

        imit_stats = policy_evaluation.eval_policy(sqil_trainer.policy, venv)

    stats = _collect_stats(imit_stats, expert_trajs)

    return stats


def main_console():
    observer_path = pathlib.Path.cwd() / "output" / "sacred" / "train_imitation"
    observer = FileStorageObserver(observer_path)
    train_imitation_ex.observers.append(observer)
    train_imitation_ex.run_commandline()


if __name__ == "__main__":
    main_console()



"""Hypothesis strategies for generating sequences of trajectories for testing."""
from typing import Dict, List, Union

import gymnasium as gym
import numpy as np
from hypothesis import strategies as st

from imitation.data import types

gym_spaces = st.sampled_from(
    [
        gym.spaces.Discrete(3),
        gym.spaces.MultiDiscrete([3, 4]),
        gym.spaces.Box(-1, 1, shape=(1,)),
        gym.spaces.Box(-1, 1, shape=(2,)),
        gym.spaces.Box(-np.inf, np.inf, shape=(2,)),
    ],
)
"""A strategy to generate spaces supported by trajectory serialization."""

info_dict_contents = st.dictionaries(
    keys=st.text(),
    values=st.one_of(
        st.integers(),
        st.floats(allow_nan=False),
        st.text(),
        st.lists(st.floats(allow_nan=False)),
    ),
)
"""A strategy to generate contents of the info dict for a trajectory."""

trajectory_length = st.integers(min_value=1, max_value=10)
"""The length of a trajectories we want to test."""


def _build_trajectory_without_reward(
    obs_space: gym.Space,
    act_space: gym.Space,
    length: int,
    info_dict_contents: Dict[str, Union[int, str, float, List[float]]],
    terminal: bool,
):
    return types.Trajectory(
        obs=np.array([obs_space.sample() for _ in range(length + 1)]),
        acts=np.array([act_space.sample() for _ in range(length)]),
        infos=np.array([info_dict_contents for _ in range(length)]),
        terminal=terminal,
    )


def _build_trajectory_with_rew(
    obs_space: gym.Space,
    act_space: gym.Space,
    length: int,
    info_dict_contents: Dict[str, Union[int, str, float, List[float]]],
    terminal: bool,
    min_rew: float,
    max_rew: float,
):
    return types.TrajectoryWithRew(
        obs=np.array([obs_space.sample() for _ in range(length + 1)]),
        acts=np.array([act_space.sample() for _ in range(length)]),
        infos=np.array([info_dict_contents for _ in range(length)]),
        terminal=terminal,
        rews=np.random.uniform(min_rew, max_rew, size=length),
    )


trajectory = st.one_of(
    st.builds(
        _build_trajectory_without_reward,
        obs_space=gym_spaces,
        act_space=gym_spaces,
        length=trajectory_length,
        info_dict_contents=info_dict_contents,
        terminal=st.booleans(),
    ),
    st.builds(
        _build_trajectory_with_rew,
        obs_space=gym_spaces,
        act_space=gym_spaces,
        length=trajectory_length,
        info_dict_contents=info_dict_contents,
        terminal=st.booleans(),
        min_rew=st.floats(min_value=-100, max_value=100),
        max_rew=st.floats(min_value=-100, max_value=100),
    ),
)
"""A strategy to generate a single trajectory (with or without reward) for testing."""


# Note: those shared strategies are used to ensure that each trajectory in a list
# is generated using the same spaces.
_shared_obs_space = st.shared(gym_spaces, key="obs_space")
_shared_act_space = st.shared(gym_spaces, key="act_space")

trajectories_without_reward_list = st.lists(
    st.builds(
        _build_trajectory_without_reward,
        obs_space=_shared_obs_space,
        act_space=_shared_act_space,
        length=trajectory_length,
        info_dict_contents=info_dict_contents,
        terminal=st.booleans(),
    ),
    min_size=1,
    max_size=10,
)
"""A strategy to generate lists of trajectories (without reward) for testing.

All trajectories in the list are generated using the same spaces.
"""

trajectories_with_reward_list = st.lists(
    st.builds(
        _build_trajectory_with_rew,
        obs_space=_shared_obs_space,
        act_space=_shared_act_space,
        length=trajectory_length,
        info_dict_contents=info_dict_contents,
        terminal=st.booleans(),
        min_rew=st.floats(min_value=-100, max_value=100),
        max_rew=st.floats(min_value=-100, max_value=100),
    ),
    min_size=1,
    max_size=10,
)
"""A strategy to generate lists of trajectories (with reward) for testing.

All trajectories in the list are generated using the same spaces.
"""

trajectories_list = st.one_of(
    trajectories_without_reward_list,
    trajectories_with_reward_list,
)
"""A strategy to generate lists of trajectories (with or without reward) for testing.

All trajectories in the list are generated using the same spaces.
They either all have reward or none of them have reward.
"""
